"""
Deprecated Final Exp Circuits since we now use the final exp witness.
Kept for reference/in case we need it in the future.
"""

import copy
from random import randint

from garaga.definitions import BLS12_381_ID, BN254_ID, CURVES, CurveID, get_sparsity
from garaga.extension_field_modulo_circuit import (
    AccPolyInstructionType,
    ExtensionFieldModuloCircuit,
    ModuloCircuitElement,
    Polynomial,
    PyFelt,
)
from garaga.hints.extf_mul import (
    nondeterministic_extension_field_mul_divmod,
    nondeterministic_square_torus,
)
from garaga.hints.frobenius import get_frobenius_maps, get_V_torus_powers
from garaga.modulo_circuit import WriteOps
from garaga.poseidon_transcript import CairoPoseidonTranscript


class FinalExpTorusCircuit(ExtensionFieldModuloCircuit):
    def __init__(
        self, name: str, curve_id: int, extension_degree: int, hash_input: bool = True
    ):
        super().__init__(
            name=name,
            curve_id=curve_id,
            extension_degree=extension_degree,
            hash_input=hash_input,
        )
        self.frobenius_maps = {}
        self.v_torus_powers_inv = {}
        for i in [1, 2, 3]:
            _, self.frobenius_maps[i] = get_frobenius_maps(
                curve_id=curve_id, extension_degree=extension_degree, frob_power=i
            )
            self.v_torus_powers_inv[i] = get_V_torus_powers(
                curve_id, extension_degree, i
            ).get_coeffs()

            # Write to circuit. Note : add_constant will return existing circuit constant if it already exists.
            self.v_torus_powers_inv[i] = [
                self.set_or_get_constant(v) for v in self.v_torus_powers_inv[i]
            ]
        self.ops_counter.update({"MUL_TORUS": 0, "SQUARE_TORUS": 0})
        self.set_or_get_constant(self.field(-1))

    def final_exp_part1(self, X: list[PyFelt]) -> list[ModuloCircuitElement]:
        return NotImplementedError

    def square_torus(
        self: ExtensionFieldModuloCircuit, X: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Computes the square of an element X in Torus Based Arithmetic
        SQ = 1/2(X+ v/X) <=> v = X (2SQ - X)
        Use hint to avoid v/X inversion.
        From SQ and X, compute v = X (2SQ - X).
        """
        self.ops_counter["SQUARE_TORUS"] += 1
        SQ: list[PyFelt] = nondeterministic_square_torus(
            X, self.curve_id, biject_from_direct=True
        )
        SQ = self.write_elements(SQ, WriteOps.COMMIT)
        two_SQ = self.vector_add(SQ, SQ)
        two_SQ_min_X = self.vector_sub(two_SQ, X)

        # %{
        Q, V = nondeterministic_extension_field_mul_divmod(
            [X, two_SQ_min_X], self.curve_id, self.extension_degree
        )
        Q = Polynomial(Q)
        # Sanity check : ensure V is indeed V(x) = 1*x.
        assert all([v.value == 0 for i, v in enumerate(V) if i != 1])
        assert V[1].value == 1, f"V(x) = {V[1].value}"
        # %}

        # Hacky way to pass SQ as R so that it is hashed.
        # Result is known in advance to be V, and not SQ.
        # R_sparsity encoded as this bypasses the usage of R coefficients.
        # Essentially, r_sparsity encodes R=V direcly by its static sparsity.

        self.accumulate_poly_instructions[0].append(
            AccPolyInstructionType.SQUARE_TORUS,
            Pis=[X, two_SQ_min_X],
            Q=Q,
            R=SQ,
            r_sparsity=[0, 2] + [0] * (self.extension_degree - 2),
        )

        return SQ

    def n_square_torus(self, X: list[PyFelt], n: int) -> list[PyFelt]:
        result = self.square_torus(X)
        for _ in range(n - 1):
            result = self.square_torus(result)
        return result

    def mul_torus(
        self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Computes Mul(X,Y) = (X*Y + v)/(X+Y)
        """
        self.ops_counter["MUL_TORUS"] += 1
        xy = self.extf_mul([X, Y], self.extension_degree)

        num = copy.deepcopy(xy)
        num[1] = self.add(xy[1], self.set_or_get_constant(1))

        den = self.vector_add(X, Y)
        return self.extf_div(num, den, self.extension_degree)

    def inverse_torus(self, X: list[ModuloCircuitElement]):
        return [self.neg(x) for x in X]

    def decompress_torus(
        self, X: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Returns (X + w) / (X - w). Size is doubled.
        """
        zero = self.set_or_get_constant(0)
        num = [
            X[0],
            self.set_or_get_constant(1),
            X[1],
            zero,
            X[2],
            zero,
            X[3],
            zero,
            X[4],
            zero,
            X[5],
            zero,
        ]
        den = num.copy()
        den[1] = self.set_or_get_constant(-1)

        return self.extf_div(num, den, 2 * self.extension_degree, acc_index=1)

    def frobenius_torus(
        self, X: list[ModuloCircuitElement], frob_power: int
    ) -> list[ModuloCircuitElement]:
        frob = [None] * self.extension_degree
        for i, list_op in enumerate(self.frobenius_maps[frob_power]):
            list_op_result = []
            for index, constant in list_op:
                if constant == 1:
                    list_op_result.append(X[index])
                else:
                    list_op_result.append(
                        self.mul(
                            X[index], self.set_or_get_constant(self.field(constant))
                        )
                    )
            frob[i] = list_op_result[0]
            for op_res in list_op_result[1:]:
                frob[i] = self.add(frob[i], op_res)

        if len(self.v_torus_powers_inv[frob_power]) == 1:
            return self.vector_scale(frob, self.v_torus_powers_inv[frob_power][0])
        else:
            Y = [
                self.set_or_get_constant(v) for v in self.v_torus_powers_inv[frob_power]
            ]
            return self.extf_mul(
                Ps=[frob, Y],
                extension_degree=self.extension_degree,
                Ps_sparsities=[None, get_sparsity(Y)],
            )

    def easy_part(
        self, num: list[PyFelt], den: list[PyFelt]
    ) -> list[ModuloCircuitElement]:
        """
        Computes the easy part of the final exponentiation.
        """

        assert len(num) == 6, f"Expected 6 elements in num, got {len(num)}"
        assert len(den) == 6, f"Expected 6 elements in den, got {len(den)}"

        num = self.write_elements(num, operation=WriteOps.INPUT)
        num = self.vector_neg(num)
        den = self.write_elements(den, WriteOps.INPUT)

        c = self.extf_div(num, den, self.extension_degree)
        t0 = self.frobenius_torus(c, 2)
        c = self.mul_torus(t0, c)
        return c

    def final_exp_finalize(self, t0: list[PyFelt], t2: list[PyFelt]):
        # Computes Decompress Torus(MulTorus(t0, t2)).
        # Only valid if (t0 + t2) != 0.
        t0 = self.write_elements(t0, WriteOps.INPUT)
        t2 = self.write_elements(t2, WriteOps.INPUT)

        # for x in t0:
        #     print(f"T0 input {hex(x.value)}")
        # for x in t2:
        #     print(f"T2 input {hex(x.value)}")

        mul = self.mul_torus(t0, t2)

        res = self.decompress_torus(mul)
        self.finalize_circuit()
        return res


class GaragaBLS12_381FinalExp(FinalExpTorusCircuit):
    def __init__(
        self, name: str = None, hash_input: bool = True, init_hash: int = None
    ):
        super().__init__(
            name=name or "Final Exp BLS12_381",
            curve_id=BLS12_381_ID,
            extension_degree=6,
            hash_input=hash_input,
        )
        if init_hash is not None:
            self.transcript = CairoPoseidonTranscript(init_hash)

    def expt_half_torus(self, X: list[ModuloCircuitElement]):
        z = self.square_torus(X)
        z = self.mul_torus(X, z)
        z = self.square_torus(z)
        z = self.square_torus(z)
        z = self.mul_torus(X, z)
        z = self.n_square_torus(z, 3)
        z = self.mul_torus(X, z)
        z = self.n_square_torus(z, 9)
        z = self.mul_torus(X, z)
        z = self.n_square_torus(z, 32)
        z = self.mul_torus(X, z)
        z = self.n_square_torus(z, 15)
        z = self.inverse_torus(z)
        return z

    def expt_torus(self, X):
        z = self.expt_half_torus(X)
        return self.square_torus(z)

    def final_exp_part1(self, num: list[PyFelt], den: list[PyFelt]) -> list[PyFelt]:
        """
        X is a list of 12 elements in the tower of extension fields.
        """
        c = self.easy_part(num, den)

        # 2. Hard part (up to permutation)
        # 3(p⁴-p²+1)/r
        # Daiki Hayashida, Kenichiro Hayasaka and Tadanori Teruya
        # https://eprint.iacr.org/2020/875.pdf
        # performed in torus compressed form
        t0_0 = self.square_torus(c)
        t1 = self.expt_half_torus(t0_0)
        t2 = self.inverse_torus(c)
        t1 = self.mul_torus(t1, t2)
        t2 = self.expt_torus(t1)
        t1 = self.inverse_torus(t1)
        t1 = self.mul_torus(t1, t2)
        t2 = self.expt_torus(t1)
        t1 = self.frobenius_torus(t1, 1)
        t1 = self.mul_torus(t1, t2)
        t0 = self.expt_torus(t1)
        t2 = self.expt_torus(t0)
        t0 = self.frobenius_torus(t1, 2)
        t1 = self.inverse_torus(t1)
        t1 = self.mul_torus(t1, t2)

        # Output in fixed order (t0, c, _sum):
        t1 = self.mul_torus(t1, t0)
        self.extend_output(t1)
        c = self.mul_torus(c, t0_0)
        self.extend_output(c)
        # The final exp result is DecompressTorus(MulTorus(c, t1)
        # MulTorus(t1, c) = (t1*c + v)/(t1 + c).
        # (t1+c = 0) ==> MulTorus(t1, c) is E12.One() in the Torus.
        _sum = self.vector_add(t1, c)
        self.extend_output(_sum)
        # From this case we can conclude the result is 1 or !=1 without decompression.
        # In case we want to decompress to get the result in GT,
        # we might need to decompress with another circuit, if the result is not 1 (_sum!=0).
        return t1, c, _sum


class GaragaBN254FinalExp(FinalExpTorusCircuit):
    def __init__(
        self, name: str = None, hash_input: bool = True, init_hash: int = None
    ):
        super().__init__(
            name=name or "Final Exp BN254",
            curve_id=BN254_ID,
            extension_degree=6,
            hash_input=hash_input,
        )
        if init_hash is not None:
            self.transcript = CairoPoseidonTranscript(init_hash)

    def expt_torus(self, X: list[PyFelt]):
        t3 = self.square_torus(X)
        t5 = self.square_torus(t3)
        result = self.square_torus(t5)
        t0 = self.square_torus(result)
        t2 = self.mul_torus(X, t0)
        t0 = self.mul_torus(t3, t2)
        t1 = self.mul_torus(X, t0)
        t4 = self.mul_torus(result, t2)
        t6 = self.square_torus(t2)
        t1 = self.mul_torus(t0, t1)
        t0 = self.mul_torus(t3, t1)
        t6 = self.n_square_torus(t6, 6)
        t5 = self.mul_torus(t5, t6)
        t5 = self.mul_torus(t4, t5)
        t5 = self.n_square_torus(t5, 7)
        t4 = self.mul_torus(t4, t5)
        t4 = self.n_square_torus(t4, 8)
        t4 = self.mul_torus(t0, t4)
        t3 = self.mul_torus(t3, t4)
        t3 = self.n_square_torus(t3, 6)
        t2 = self.mul_torus(t2, t3)
        t2 = self.n_square_torus(t2, 8)
        t2 = self.mul_torus(t0, t2)
        t2 = self.n_square_torus(t2, 6)
        t2 = self.mul_torus(t0, t2)
        t2 = self.n_square_torus(t2, 10)
        t1 = self.mul_torus(t1, t2)
        t1 = self.n_square_torus(t1, 6)
        t0 = self.mul_torus(t0, t1)
        z = self.mul_torus(result, t0)
        return z

    def final_exp_part1(
        self, num: list[PyFelt], den: list[PyFelt]
    ) -> list[ModuloCircuitElement]:
        """
        single pairing -> unsafe = True = 1
        double pairing -> unsafe = False = 0
        """
        c = self.easy_part(num, den)

        # 2. Hard part (up to permutation)
        # 2x₀(6x₀²+3x₀+1)(p⁴-p²+1)/r
        # Duquesne and Ghammam
        # https://eprint.iacr.org/2015/192.pdf
        # Fuentes et al. (alg. 6)
        # performed in torus compressed form
        t0 = self.expt_torus(c)
        t0 = self.inverse_torus(t0)
        t0 = self.square_torus(t0)
        t1 = self.square_torus(t0)
        t1 = self.mul_torus(t0, t1)
        t2 = self.expt_torus(t1)
        t2 = self.inverse_torus(t2)
        t3 = self.inverse_torus(t1)
        t1 = self.mul_torus(t2, t3)
        t3 = self.square_torus(t2)
        t4 = self.expt_torus(t3)
        t4 = self.mul_torus(t1, t4)
        t3 = self.mul_torus(t0, t4)
        t0 = self.mul_torus(t2, t4)
        t0 = self.mul_torus(c, t0)
        t2 = self.frobenius_torus(t3, 1)
        t0 = self.mul_torus(t2, t0)
        t2_tmp = self.frobenius_torus(t4, 2)
        t2 = self.inverse_torus(c)
        t2 = self.mul_torus(t2, t3)

        # Output in fixed order (t0, t2, _sum):
        t0 = self.mul_torus(t0, t2_tmp)
        self.extend_output(t0)
        t2 = self.frobenius_torus(t2, 3)
        self.extend_output(t2)
        # The final exp result is DecompressTorus(MulTorus(t0, t2)).
        # MulTorus(t0, t2) = (t0*t2 + v)/(t0 + t2).
        # (T0+T2 = 0) ==> MulTorus(t0, t2) is one in the Torus.
        _sum = self.vector_add(t0, t2)
        self.extend_output(_sum)
        # From this case we can conclude the result is 1 or !=1 without decompression.
        # In case we want to decompress to get the result in GT,
        # we might need to decompress with another circuit, if the result is not 1 (_sum!=0).

        return t0, t2, _sum


GaragaFinalExp = {
    CurveID.BN254: GaragaBN254FinalExp,
    CurveID.BLS12_381: GaragaBLS12_381FinalExp,
}


def test_final_exp(curve_id: CurveID):
    from garaga.definitions import G1G2Pair, G1Point, G2Point, tower_to_direct

    order = CURVES[curve_id.value].n
    pairs = []
    n_pairs = 1
    unsafe = True if n_pairs == 1 else False
    for _ in range(n_pairs):
        n1, n2 = randint(1, order), randint(1, order)
        p1, p2 = G1Point.get_nG(curve_id, n1), G2Point.get_nG(curve_id, n2)
        pairs.append(G1G2Pair(p1, p2))

    base_class = GaragaFinalExp[curve_id]
    part1 = base_class(hash_input=False)
    field = part1.field

    XT: list[int] = G1G2Pair.miller(pairs).value_coeffs
    ET: list[int] = G1G2Pair.pair(pairs).value_coeffs

    XT = [part1.field(x) for x in XT]
    ET = [part1.field(x) for x in ET]

    XD = tower_to_direct(XT, curve_id.value, 12)
    ED = tower_to_direct(ET, curve_id.value, 12)

    num_indexes = [0, 2, 4, 6, 8, 10]
    den_indexes = [1, 3, 5, 7, 9, 11]
    num = [XD[i] for i in num_indexes]

    if unsafe:
        den = [XD[i] for i in den_indexes]
    else:
        if [x.value for x in [XD[i] for i in den_indexes]] == [0, 0, 0, 0, 0, 0]:
            den = [
                field.one(),
                field.zero(),
                field.zero(),
                field.zero(),
                field.zero(),
                field.zero(),
            ]

        else:
            den = [XD[i] for i in den_indexes]

    t0, t2, _sum = part1.final_exp_part1(num, den)
    part1.finalize_circuit()
    _sum = [x.value for x in _sum]

    part2 = base_class(hash_input=False)
    if _sum == [0, 0, 0, 0, 0, 0]:
        f = [field.one()] + [field.zero] * 11
    else:
        t0 = [x.felt for x in t0]
        t2 = [x.felt for x in t2]
        f = part2.final_exp_finalize(t0, t2)

    assert [f.value for f in f] == [
        e.value for e in ED
    ], f"Final exp in circuit and internal do not match f={[f.value for f in f]}\ne={[e.value for e in ED]}"
    # print(f"{curve_id} Final Exp random test pass")
    return part1, part2


if __name__ == "__main__":
    pass

    from garaga.definitions import CurveID, Polynomial

    # def test_frobenius_torus():
    #     from archive_tmp.bn254.pairing_final_exp import frobenius_torus
    #     field = get_base_field(CurveID.BN254.value)
    #     X = [field(random.randint(0, field.p - 1)) for _ in range(6)]
    #     t = FinalExpTorusCircuit("test", CurveID.BN254.value, 6)
    #     t.create_powers_of_Z(field(2))
    #     X = t.write_elements(X)
    #     XF = t.frobenius_torus(X, 1)
    #     # Xpoly = Polynomial([x.felt for x in X])
    #     # XFpoly = Xpoly.pow(field.p, get_irreducible_poly(CurveID.BN254.value, ))
    #     # assert t.finalize_circuit()
    #     # t.values_segment = t.values_segment.non_interactive_transform()
    #     TT = frobenius_torus([x.value for x in X])
    #     assert all(x.value == y for x, y in zip(XF, TT))
    #     t.print_value_segment()
    # test_final_exp(CurveID.BN254)
    # test_final_exp(CurveID.BLS12_381)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar

from garaga.algebra import FunctionFelt, ModuloCircuitElement, PyFelt
from garaga.definitions import STARK, G1Point, G2Point, get_base_field
from garaga.hints import io
from garaga.hints.io import (
    int_array_to_u384_array,
    int_to_u2XX,
    int_to_u256,
    int_to_u384,
)

T = TypeVar("T", bound="Cairo1SerializableStruct")

from enum import Enum


class CairoOption(Enum):
    SOME = 0
    NONE = 1
    VOID = 2  # Special case to serialize nothing at all.


@dataclass(slots=True)
class Cairo1SerializableStruct(ABC):
    name: str
    elmts: list[ModuloCircuitElement | PyFelt | "Cairo1SerializableStruct"]

    def __post_init__(self):
        assert type(self.name) == str
        if isinstance(self.elmts, (list, tuple)):
            if len(self.elmts) > 0:
                if isinstance(self.elmts[0], Cairo1SerializableStruct):
                    assert all(
                        isinstance(elmt, self.elmts[0].__class__) for elmt in self.elmts
                    ), f"All elements of {self.name} must be of the same type"

                else:
                    assert all(
                        isinstance(elmt, (ModuloCircuitElement, PyFelt))
                        for elmt in self.elmts
                    ), f"All elements of {self.name} must be of type ModuloCircuitElement or PyFelt, got {type(self.elmts[0])}"
        else:
            assert self.elmts is None, f"Elmts must be a list or None, got {self.elmts}"

    @property
    def struct_name(self) -> str:
        return self.__class__.__name__

    @property
    def bits(self) -> int:
        return self.elmts[0].p.bit_length()

    def serialize_input_signature(self) -> str:
        return f"{self.name}:{self.struct_name}"

    @abstractmethod
    def serialize(self, raw: bool = False) -> str:
        pass

    @abstractmethod
    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        pass

    @abstractmethod
    def dump_to_circuit_input(self) -> str:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    def _serialize_to_calldata(self) -> list[int]:
        pass

    def serialize_to_calldata(self, *args, **kwargs) -> list[int]:
        data = self._serialize_to_calldata(*args, **kwargs)
        # print(
        #     f"\nSerializing {self.name} {self.struct_name} to calldata: {data} {len(data)}"
        # )
        return data


class StructArray(Cairo1SerializableStruct, Generic[T]):
    elmts: list[T]

    @property
    def struct_name(self) -> str:
        return "Array<" + self.elmts[0].struct_name + ">"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for struct in self.elmts:
            code += struct.dump_to_circuit_input()
        return code

    def __len__(self) -> int:
        return sum(len(struct) for struct in self.elmts)

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def serialize(self, raw: bool = False, const: bool = False) -> str:
        if const:
            raw_struct = "["
        else:
            raw_struct = "array!["

        for struct in self.elmts:
            raw_struct += struct.serialize(raw=True) + ","
        raw_struct += "]\n"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def serialize_to_calldata(self) -> list[int]:
        cd = [len(self)]
        for elmt in self.elmts:
            cd.extend(elmt.serialize_to_calldata())
        return cd


class Struct(Cairo1SerializableStruct):
    elmts: list[Cairo1SerializableStruct]

    def __init__(
        self, struct_name: str, name: str, elmts: list[Cairo1SerializableStruct]
    ):
        super().__init__(name, elmts)
        self._struct_name = struct_name

    @property
    def struct_name(self) -> str:
        return self._struct_name

    def __post_init__(self):
        assert all(isinstance(elmt, Cairo1SerializableStruct) for elmt in self.elmts)

    def dump_to_circuit_input(self) -> str:
        return NotImplementedError

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def serialize(self, raw: bool = False) -> str:
        if raw:
            code = f"{self.struct_name} {{"
        else:
            code = f"let {self.name} = {self.struct_name} {{"
        for struct in self.elmts:
            code += f"{struct.name}: {struct.serialize(raw=True)},"
        if raw:
            code += "}"
        else:
            code += "};"
        return code

    def __len__(self) -> int:
        return sum(len(elmt) for elmt in self.elmts)

    def _serialize_to_calldata(self) -> list[int]:
        cd = []
        for elmt in self.elmts:
            cd.extend(elmt.serialize_to_calldata())
        return cd


class StructSpan(Cairo1SerializableStruct, Generic[T]):
    elmts: list[T]

    @property
    def struct_name(self) -> str:
        return "Span<" + self.elmts[0].struct_name + ">"

    def dump_to_circuit_input(self) -> str:
        code = f"let mut {self.name} = {self.name};\n"
        code += f"for val in {self.name} {{\n"

        self.elmts[0].name = "*val"
        code += self.elmts[0].dump_to_circuit_input()
        code += "};\n"
        return code

    def __len__(self) -> int:
        return sum(len(struct) for struct in self.elmts)

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def serialize(self, raw: bool = False, is_option: bool = False) -> str:
        if self.elmts is None:
            raw_struct = "Option::None"
            if raw:
                return raw_struct
            else:
                return f"let {self.name} = {raw_struct};\n"
        else:
            raw_struct = "array!["
            for struct in self.elmts:
                raw_struct += struct.serialize(raw=True) + ","
            raw_struct += "].span()"
            if is_option:
                raw_struct = f"Option::Some({raw_struct})"
            if raw:
                return raw_struct
            else:
                return f"let {self.name} = {raw_struct};\n"

    def _serialize_to_calldata(
        self, option: CairoOption = None, as_pure_felt252_array: bool = False
    ) -> list[int]:
        cd = []
        if option:
            if option == CairoOption.SOME:
                cd.append(0)
            elif option == CairoOption.NONE:
                cd.append(1)
                return cd
            elif option == CairoOption.VOID:
                return cd
            else:
                raise ValueError(f"Invalid option: {option}")

        cd.append(len(self.elmts))
        for elmt in self.elmts:
            cd.extend(elmt._serialize_to_calldata())

        if as_pure_felt252_array:
            cd[0] = len(cd) - 1
        return cd


# class u288(Cairo1SerializableStruct):
#     def serialize(self, raw: bool = False) -> str:
#         assert len(self.elmts) == 1
#         raw_struct = f"{int_to_u288(self.elmts[0].value)}"
#         if raw:
#             return raw_struct
#         else:
#             return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

#     def _serialize_to_calldata(self) -> list[int]:
#         assert len(self.elmts) == 1
#         return io.bigint_split_array(self.elmts, n_limbs=3, prepend_length=False)

#     def extract_from_circuit_output(
#         self, offset_to_reference_map: dict[int, str]
#     ) -> str:
#         assert len(self.elmts) == 1
#         return f"let {self.name}:{self.struct_name} = outputs.get_output({offset_to_reference_map[self.elmts[0].offset]});"

#     def dump_to_circuit_input(self) -> str:
#         return f"circuit_inputs = circuit_inputs.next_u288({self.name});\n"

#     def __len__(self) -> int:
#         if self.elmts is not None:
#             assert len(self.elmts) == 1
#             return 1
#         else:
#             return 1


class u384(Cairo1SerializableStruct):
    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 1
        raw_struct = f"{int_to_u384(self.elmts[0].value)}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        assert len(self.elmts) == 1
        return io.bigint_split_array(self.elmts, prepend_length=False)

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 1
        return f"let {self.name}:{self.struct_name} = outputs.get_output({offset_to_reference_map[self.elmts[0].offset]});"

    def dump_to_circuit_input(self) -> str:
        return f"circuit_inputs = circuit_inputs.next_2({self.name});\n"

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 1
            return 1
        else:
            return 1


class GenericT(Cairo1SerializableStruct):
    # <T> => u384 or u288

    @property
    def struct_name(self) -> str:
        return "T"

    def serialize(self, raw: bool = False) -> str:
        if self.bits <= 288:
            curve_id = 0
        else:
            curve_id = 1
        assert len(self.elmts) == 1
        raw_struct = f"{int_to_u2XX(self.elmts[0].value, curve_id)}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        assert len(self.elmts) == 1
        if self.bits <= 288:
            return io.bigint_split_array(self.elmts, prepend_length=False)
        else:
            return io.bigint_split_array(self.elmts, n_limbs=3, prepend_length=False)

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def dump_to_circuit_input(self) -> str:
        return f"circuit_inputs = circuit_inputs.next_2({self.name});\n"

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 1
            return 1
        else:
            return 1


class u256(Cairo1SerializableStruct):
    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 1
        raw_struct = f"{int_to_u256(self.elmts[0].value)}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        assert len(self.elmts) == 1
        return list(io.split_128(self.elmts[0].value))

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def dump_to_circuit_input(self) -> str:
        raise NotImplementedError

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 1
            return 1
        else:
            return 1


class u256Span(Cairo1SerializableStruct):
    @property
    def struct_name(self) -> str:
        return "Span<u256>"

    def serialize(self, raw: bool = False) -> str:
        raw_struct = f"{io.int_array_to_u256_array(self.elmts)}.span()"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        return io.bigint_split_array(
            self.elmts, n_limbs=2, base=2**128, prepend_length=True
        )

    def dump_to_circuit_input(self) -> str:
        code = f"""
    for val in {self.name} {{
        circuit_inputs = circuit_inputs.next_u256(*val);
    }};"""
        return code

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def __len__(self) -> int:
        if self.elmts is not None:
            return len(self.elmts)
        else:
            return None


class u128(Cairo1SerializableStruct):
    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 1
        assert 0 <= self.elmts[0].value < 2**128
        raw_struct = f"{hex(self.elmts[0].value)}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        assert len(self.elmts) == 1
        assert 0 <= self.elmts[0].value < 2**128
        return [self.elmts[0].value]

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def dump_to_circuit_input(self) -> str:
        return f"circuit_inputs = circuit_inputs.next_u128({self.name});\n"

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 1
            return 1
        else:
            return 1


class u128Span(Cairo1SerializableStruct):
    @property
    def struct_name(self) -> str:
        return "Span<u128>"

    def serialize(self, raw: bool = False) -> str:
        raw_struct = f"{io.int_array_to_u128_array(self.elmts)}.span()"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        return io.bigint_split_array(
            self.elmts, n_limbs=1, base=2**128, prepend_length=True
        )

    def dump_to_circuit_input(self) -> str:
        code = f"""
    for val in {self.name} {{
        circuit_inputs = circuit_inputs.next_u128(*val);
    }};"""
        return code

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def __len__(self) -> int:
        if self.elmts is not None:
            return len(self.elmts)
        else:
            return None


class Tuple(Cairo1SerializableStruct):
    elmts: list[Cairo1SerializableStruct]

    def __init__(self, name: str, elmts: list[Cairo1SerializableStruct]):
        super().__init__(name, elmts)
        self.members_names = [member.name for member in elmts]

    @property
    def struct_name(self) -> str:
        return f"({','.join([elmt.struct_name for elmt in self.elmts])})"

    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) > 0
        raw_struct = f"({','.join([elmt.serialize(raw=True) for elmt in self.elmts])})"
        if raw:
            return raw_struct
        else:
            return f"let ({','.join([elmt.name for elmt in self.elmts])}):{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        cd = []
        for elmt in self.elmts:
            cd.extend(elmt._serialize_to_calldata())
        return cd

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def dump_to_circuit_input(self) -> str:
        code = ""
        # Need to unpack the tuple  :
        code = f"let ({','.join([self.members_names[i] for i in range(len(self.elmts))])}) = {self.name};\n"
        # Now call dump_to_circuit_input for each member
        # Names must correspond to the members names
        for i, elmt in enumerate(self.elmts):
            assert (
                elmt.name == self.members_names[i]
            ), f"Tuple member {i} has name {elmt.name} instead of {self.members_names[i]}"
            code += elmt.dump_to_circuit_input()
        return code

    def __len__(self) -> int:
        return sum(len(elmt) for elmt in self.elmts)


class felt252(Cairo1SerializableStruct):
    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 1
        raw_struct = f"{hex(self.elmts[0].value)}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        assert len(self.elmts) == 1
        return [self.elmts[0].value % STARK]

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError

    def dump_to_circuit_input(self) -> str:
        code = f"let {self.name}_384:u384 = {self.name}.into();\n"
        code += f"circuit_inputs = circuit_inputs.next_2({self.name}_384);\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 1
            return 1
        else:
            return 1


class u384Array(Cairo1SerializableStruct):
    def serialize(self, raw: bool = False) -> str:
        if len(self.elmts) == 0:
            raw_struct = "array![]"
        else:
            bits = self.bits
            if self.name == "g_rhs_sqrt":
                # Temp fix before we change the MSMHint and G1Points to depend on the curve id
                # Todo : remove this
                curve_id = 1
            else:
                curve_id = 0 if bits <= 288 else 1

            raw_struct = f"{io.int_array_to_u2XX_array(self.elmts, curve_id=curve_id)}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        if len(self.elmts) == 0:
            return [0]
        bits = self.bits
        if bits <= 288 and self.name != "g_rhs_sqrt":
            return io.bigint_split_array(self.elmts, n_limbs=3, prepend_length=True)
        else:
            return io.bigint_split_array(self.elmts, n_limbs=4, prepend_length=True)

    @property
    def struct_name(self) -> str:
        bits = self.bits
        if bits <= 288:
            return "Array<u288>"
        elif bits <= 384:
            return "Array<u384>"
        else:
            raise ValueError(f"Unsupported bit length for u384Array: {bits}")

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        return f"let {self.name} = array![{','.join([f'outputs.get_output({offset_to_reference_map[elmt.offset]})' for elmt in self.elmts])}];"

    def dump_to_circuit_input(self) -> str:
        next_fn = "next_2"
        code = f"""
    for val in {self.name}.span() {{
        circuit_inputs = circuit_inputs.{next_fn}(*val);
    }};"""
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            return len(self.elmts)
        else:
            return None


class FunctionFeltCircuit(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[u384Span]):
        super().__init__(name, elmts)
        self.members_names = ("a_num", "a_den", "b_num", "b_den")

    @property
    def a_num(self) -> list[ModuloCircuitElement | PyFelt]:
        return self.elmts[0].elmts

    @property
    def a_den(self) -> list[ModuloCircuitElement | PyFelt]:
        return self.elmts[1].elmts

    @property
    def b_num(self) -> list[ModuloCircuitElement | PyFelt]:
        return self.elmts[2].elmts

    @property
    def b_den(self) -> list[ModuloCircuitElement | PyFelt]:
        return self.elmts[3].elmts

    @property
    def struct_name(self) -> str:
        return "FunctionFelt"

    @staticmethod
    def from_FunctionFelt(
        name: str, f: FunctionFelt, msm_size: int, batched: bool = False
    ) -> "FunctionFeltCircuit":
        _a_num, _a_den, _b_num, _b_den = io.padd_function_felt(
            f, msm_size, py_felt=True, batched=batched
        )
        return FunctionFeltCircuit(
            name=name,
            elmts=[
                u384Span("a_num", _a_num),
                u384Span("a_den", _a_den),
                u384Span("b_num", _b_num),
                u384Span("b_den", _b_den),
            ],
        )

    def _serialize_to_calldata(self) -> list[int]:
        cd = []
        assert len(self.elmts) == 4
        bits = self.elmts[0].bits
        if bits <= 288:
            for elmt in self.elmts:
                cd.extend(
                    io.bigint_split_array(elmt.elmts, n_limbs=3, prepend_length=True)
                )
        else:
            for elmt in self.elmts:
                cd.extend(
                    io.bigint_split_array(elmt.elmts, n_limbs=4, prepend_length=True)
                )
        return cd

    def serialize_input_signature(self) -> str:
        return f"{self.name}:FunctionFelt<T>"

    def serialize(self, raw: bool = False) -> str:
        raw_struct = f"FunctionFelt {{ {','.join([f'{self.members_names[i]}: {self.elmts[i].serialize(raw=True)}' for i in range(4)])} }}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"""
            for val in {self.name}.{mem_name} {{
                circuit_inputs = circuit_inputs.next_2(*val);
            }};
            """
        return code

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: {self.elmts[i].extract_from_circuit_output(offset_to_reference_map)}' for i in range(4)])} }};"

    def __len__(self) -> int:
        return sum(len(elmt) for elmt in self.elmts)


class u384Span(Cairo1SerializableStruct):
    def serialize(self, raw: bool = False) -> str:
        raw_struct = f"{int_array_to_u384_array(self.elmts)}.span()"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        return io.bigint_split_array(self.elmts, prepend_length=True)

    @property
    def struct_name(self) -> str:
        return "Span<u384>"

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        return f"let {self.name}:{self.struct_name} = array![{','.join([f'outputs.get_output({offset_to_reference_map[elmt.offset]})' for elmt in self.elmts])}].span();"

    def dump_to_circuit_input(self) -> str:
        code = f"""let mut {self.name} = {self.name};
    for val in {self.name} {{
        circuit_inputs = circuit_inputs.next_2(*val);
    }};
    """
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            return len(self.elmts)
        else:
            return None


class BLSProcessedPair(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("yInv", "xNegOverY")

    @property
    def struct_name(self) -> str:
        return "BLSProcessedPair"

    def serialize(self) -> str:
        assert len(self.elmts) == 2
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{yInv: {int_to_u384(self.elmts[0].value)}, xNegOverY: {int_to_u384(self.elmts[1].value)}}};"

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 2
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(2)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 2
            return 2
        else:
            return 2


class BNProcessedPair(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("yInv", "xNegOverY", "QyNeg0", "QyNeg1")

    @property
    def struct_name(self) -> str:
        return "BNProcessedPair"

    def serialize(self) -> str:
        assert len(self.elmts) == 4
        members = ", ".join(
            [
                f"{name}: {int_to_u384(self.elmts[i].value)}"
                for i, name in enumerate(self.members_names)
            ]
        )
        return (
            f"let {self.name}:{self.struct_name} = {self.struct_name} {{{members}}};\n"
        )

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 4
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(4)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 4
            return 4
        else:
            return 4


class G1PointCircuit(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("x", "y")

    @staticmethod
    def from_G1Point(name: str, point: G1Point) -> "G1PointCircuit":
        field = get_base_field(point.curve_id)
        return G1PointCircuit(name=name, elmts=[field(point.x), field(point.y)])

    @property
    def struct_name(self) -> str:
        return "G1Point"

    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 2
        raw_struct = f"{self.struct_name} {{x: {int_to_u384(self.elmts[0].value)}, y: {int_to_u384(self.elmts[1].value)}}}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        return io.bigint_split_array(self.elmts, prepend_length=False)

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 2
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(2)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 2
            return 2
        else:
            return 2


class G2PointCircuit(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("x0", "x1", "y0", "y1")

    @staticmethod
    def from_G2Point(name: str, point: G2Point) -> "G2PointCircuit":
        field = get_base_field(point.curve_id)
        return G2PointCircuit(
            name=name,
            elmts=[
                field(point.x[0]),
                field(point.x[1]),
                field(point.y[0]),
                field(point.y[1]),
            ],
        )

    @property
    def struct_name(self) -> str:
        return "G2Point"

    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 4
        raw_struct = f"{self.struct_name} {{x0: {int_to_u384(self.elmts[0].value)}, x1: {int_to_u384(self.elmts[1].value)}, y0: {int_to_u384(self.elmts[2].value)}, y1: {int_to_u384(self.elmts[3].value)}}}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 4
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(4)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 4
            return 4
        else:
            return 4


class G2Line(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("r0a0", "r0a1", "r1a0", "r1a1")

    def serialize_input_signature(self) -> str:
        bits = self.bits
        if bits <= 288:
            return f"{self.name}:G2Line<u288>"
        else:
            return f"{self.name}:G2Line<u384>"

    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 4
        bits = self.bits
        if bits <= 288:
            curve_id = 0
        else:
            curve_id = 1

        raw_struct = f"{self.struct_name} {{r0a0: {int_to_u2XX(self.elmts[0].value, curve_id=curve_id)}, r0a1: {int_to_u2XX(self.elmts[1].value, curve_id=curve_id)}, r1a0: {int_to_u2XX(self.elmts[2].value, curve_id=curve_id)}, r1a1: {int_to_u2XX(self.elmts[3].value, curve_id=curve_id)}}}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.struct_name} = {raw_struct};\n"

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 4
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(4)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        next_fn = "next_2"
        for mem_name in self.members_names:
            code += (
                f"circuit_inputs = circuit_inputs.{next_fn}({self.name}.{mem_name});\n"
            )
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 4
            return 4
        else:
            return 4


class FunctionFeltEvaluations(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("a_num", "a_den", "b_num", "b_den")

    def serialize(self) -> str:
        assert len(self.elmts) == 4
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{a_num: {int_to_u384(self.elmts[0].value)}, a_den: {int_to_u384(self.elmts[1].value)}, b_num: {int_to_u384(self.elmts[2].value)}, b_den: {int_to_u384(self.elmts[3].value)}}};\n"

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 4
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(4)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 4
            return 4
        else:
            return 4


class G1G2PairCircuit(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("p.x", "p.y", "q.x0", "q.x1", "q.y0", "q.y1")

    @property
    def struct_name(self) -> str:
        return "G1G2Pair"

    def serialize(self) -> str:
        return (
            f"let {self.name}:{self.struct_name} = {self.struct_name} {{"
            f"p: G1Point{{ x:{int_to_u384(self.elmts[0].value)}, y: {int_to_u384(self.elmts[1].value)}}}, "
            f"q: G2Point{{ x0:{int_to_u384(self.elmts[2].value)}, x1: {int_to_u384(self.elmts[3].value)}, y0: {int_to_u384(self.elmts[4].value)}, y1: {int_to_u384(self.elmts[5].value)}}}}};\n"
        )

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 6
        return f"let {self.name}:{self.struct_name} = {self.struct_name} {{ {','.join([f'{self.members_names[i]}: outputs.get_output({offset_to_reference_map[self.elmts[i].offset]})' for i in range(6)])} }};"

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 6
            return 6
        else:
            return 6


class E12D(Cairo1SerializableStruct):
    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 12
        code = (
            f"let {self.name}:{self.__class__.__name__} = {self.__class__.__name__}{{\n"
        )
        for i, elmt in enumerate(self.elmts):
            code += (
                f"w{i}:outputs.get_output({offset_to_reference_map[elmt.offset]}),\n"
            )
        code += "};"
        return code

    @property
    def struct_name(self) -> str:
        p = self.elmts[0].p
        if p.bit_length() <= 288:
            return "E12D<u288>"
        else:
            return "E12D<u384>"

    def serialize(self, raw: bool = False, is_option: bool = False) -> str:
        if self.elmts is None:
            raw_struct = "Option::None"
            if raw:
                return raw_struct
            else:
                return f"let {self.name}:Option<{self.__class__.__name__}> = {raw_struct};\n"
        else:
            assert len(self.elmts) == 12
            bits: int = self.elmts[0].p.bit_length()
            if bits <= 288:
                curve_id = 0
            else:
                curve_id = 1

            raw_struct = (
                f"{self.__class__.__name__}{{"
                + f"{','.join([f'w{i}: {int_to_u2XX(self.elmts[i].value, curve_id=curve_id)}' for i in range(len(self))])}}}"
            )
            if is_option:
                raw_struct = f"Option::Some({raw_struct})"
            if raw:
                return raw_struct
            else:
                return f"let {self.name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        bits: int = self.bits
        if bits <= 288:
            return io.bigint_split_array(self.elmts, n_limbs=3, prepend_length=False)
        elif bits <= 384:
            return io.bigint_split_array(self.elmts, n_limbs=4, prepend_length=False)
        else:
            raise ValueError(f"Unsupported bit length for E12D: {bits}")

    def dump_to_circuit_input(self) -> str:
        bits: int = self.elmts[0].p.bit_length()
        code = ""
        next_fn = "next_2"
        for i in range(len(self)):
            code += f"circuit_inputs = circuit_inputs.{next_fn}({self.name}.w{i});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 12
            return 12
        else:
            return 12


class E12T(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = (
            "c0b0a0",
            "c0b0a1",
            "c0b1a0",
            "c0b1a1",
            "c0b2a0",
            "c0b2a1",
            "c1b0a0",
            "c1b0a1",
            "c1b1a0",
            "c1b1a1",
            "c1b2a0",
            "c1b2a1",
        )

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 12
        code = (
            f"let {self.name}:{self.__class__.__name__} = {self.__class__.__name__}{{\n"
        )
        for i, elmt in enumerate(self.elmts):
            code += f"{self.members_names[i]}:outputs.get_output({offset_to_reference_map[elmt.offset]}),\n"
        code += "};"
        return code

    def serialize(self, raw: bool = False, is_option: bool = False) -> str:
        if self.elmts is None:
            raw_struct = "Option::None"
            if raw:
                return raw_struct
            else:
                return f"let {self.name}:Option<{self.__class__.__name__}> = {raw_struct};\n"
        else:
            assert len(self.elmts) == 12

            raw_struct = (
                f"{self.__class__.__name__}{{"
                + f"{','.join([f'{self.members_names[i]}: {int_to_u384(self.elmts[i].value)}' for i in range(len(self))])}}}"
            )
            if is_option:
                raw_struct = f"Option::Some({raw_struct})"
            if raw:
                return raw_struct
            else:
                return f"let {self.name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        return io.bigint_split_array(self.elmts, n_limbs=4, prepend_length=False)

    def dump_to_circuit_input(self) -> str:
        code = ""
        for i in range(len(self)):
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{self.members_names[i]});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 12
            return 12
        else:
            return 12


class E12DMulQuotient(Cairo1SerializableStruct):
    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 11
        code = (
            f"let {self.name}:{self.__class__.__name__} = {self.__class__.__name__}{{\n"
        )
        for i, elmt in enumerate(self.elmts):
            code += (
                f"w{i}:outputs.get_output({offset_to_reference_map[elmt.offset]}),\n"
            )
        code += "};"
        return code

    @property
    def struct_name(self) -> str:
        p = self.elmts[0].p
        if p.bit_length() <= 288:
            return "E12DMulQuotient<u288>"
        else:
            return "E12DMulQuotient<u384>"

    def serialize(self, raw: bool = False, is_option: bool = False) -> str:
        if self.elmts is None:
            raw_struct = "Option::None"
            if raw:
                return raw_struct
            else:
                return f"let {self.name}:Option<{self.__class__.__name__}> = {raw_struct};\n"
        else:
            assert len(self.elmts) == 11
            bits: int = self.elmts[0].p.bit_length()
            if bits <= 288:
                curve_id = 0
            else:
                curve_id = 1

            raw_struct = (
                f"{self.__class__.__name__}{{"
                + f"{','.join([f'w{i}: {int_to_u2XX(self.elmts[i].value, curve_id=curve_id)}' for i in range(len(self))])}}}"
            )
            if is_option:
                raw_struct = f"Option::Some({raw_struct})"
            if raw:
                return raw_struct
            else:
                return f"let {self.name} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        bits: int = self.bits
        if bits <= 288:
            return io.bigint_split_array(self.elmts, n_limbs=3, prepend_length=False)
        elif bits <= 384:
            return io.bigint_split_array(self.elmts, n_limbs=4, prepend_length=False)
        else:
            raise ValueError(f"Unsupported bit length for E12D: {bits}")

    def dump_to_circuit_input(self) -> str:
        code = ""
        next_fn = "next_2"
        for i in range(len(self)):
            code += f"circuit_inputs = circuit_inputs.{next_fn}({self.name}.w{i});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 11
            return 11
        else:
            return 11


class MillerLoopResultScalingFactor(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("w0", "w2", "w4", "w6", "w8", "w10")

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        raise NotImplementedError("Never used in practice")

    def serialize_input_signature(self) -> str:
        bits = self.bits
        if bits <= 288:
            return f"{self.name}:MillerLoopResultScalingFactor<u288>"
        else:
            return f"{self.name}:MillerLoopResultScalingFactor<u384>"

    def dump_to_circuit_input(self) -> str:
        code = ""
        next_fn = "next_2"
        for mem_name in self.members_names:
            code += (
                f"circuit_inputs = circuit_inputs.{next_fn}({self.name}.{mem_name});\n"
            )
        return code

    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 6
        bits = self.bits
        if bits <= 288:
            curve_id = 0
        else:
            curve_id = 1
        raw_struct = f"{self.__class__.__name__}{{{','.join([f'{self.members_names[i]}: {int_to_u2XX(self.elmts[i].value, curve_id=curve_id)}' for i in range(len(self))])}}}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.__class__.__name__} = {raw_struct};\n"

    def _serialize_to_calldata(self) -> list[int]:
        bits = self.bits
        if bits <= 288:
            return io.bigint_split_array(self.elmts, n_limbs=3, prepend_length=False)
        else:
            return io.bigint_split_array(self.elmts, n_limbs=4, prepend_length=False)

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 6
            return 6
        else:
            return 6


class SlopeInterceptOutput(Cairo1SerializableStruct):
    def __init__(self, name: str, elmts: list[ModuloCircuitElement]):
        super().__init__(name, elmts)
        self.members_names = ("m_A0", "b_A0", "x_A2", "y_A2", "coeff0", "coeff2")

    def serialize(self, raw: bool = False) -> str:
        assert len(self.elmts) == 6
        raw_struct = f"{self.__class__.__name__}{{{','.join([f'{self.members_names[i]}: {int_to_u384(self.elmts[i].value)}' for i in range(len(self))])}}}"
        if raw:
            return raw_struct
        else:
            return f"let {self.name}:{self.__class__.__name__} = {raw_struct};\n"

    def extract_from_circuit_output(
        self, offset_to_reference_map: dict[int, str]
    ) -> str:
        assert len(self.elmts) == 6
        code = (
            f"let {self.name}:{self.__class__.__name__} = {self.__class__.__name__}{{\n"
        )
        for mem_name, elmt in zip(self.members_names, self.elmts):
            code += f"{mem_name}: outputs.get_output({offset_to_reference_map[elmt.offset]}),\n"
        code += "};"
        return code

    def dump_to_circuit_input(self) -> str:
        code = ""
        for mem_name in self.members_names:
            code += f"circuit_inputs = circuit_inputs.next_2({self.name}.{mem_name});\n"
        return code

    def __len__(self) -> int:
        if self.elmts is not None:
            assert len(self.elmts) == 6
            return 6
        else:
            return 6



from garaga.definitions import CURVES, CurveID
from garaga.extension_field_modulo_circuit import ModuloCircuitElement
from garaga.modulo_circuit import ModuloCircuit
from garaga.precompiled_circuits.fp2 import Fp2Circuits
from garaga.signature import get_isogeny_to_g1_map, get_isogeny_to_g2_map


class IsogenyG2(Fp2Circuits):
    def __init__(self, name: str, curve_id: int, compilation_mode: int = 0):
        super().__init__(
            name=name,
            curve_id=curve_id,
            compilation_mode=compilation_mode,
            generic_circuit=False,
        )
        self.curve = CURVES[curve_id]

    def set_consts(self):
        x_rational, y_rational = get_isogeny_to_g2_map(CurveID(self.curve_id))
        self.x_num = [
            self.set_or_get_constant(c) for c in x_rational.numerator.coefficients
        ]
        self.x_den = [
            self.set_or_get_constant(c) for c in x_rational.denominator.coefficients
        ]
        self.y_num = [
            self.set_or_get_constant(c) for c in y_rational.numerator.coefficients
        ]
        self.y_den = [
            self.set_or_get_constant(c) for c in y_rational.denominator.coefficients
        ]

    def run_isogeny(
        self, x: list[ModuloCircuitElement], y: list[ModuloCircuitElement]
    ) -> tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]:
        """
        Runs the isogeny to G2 map.
        """
        self.set_consts()

        x_affine_num = self.fp2_eval_horner(self.x_num, x, "x_num")
        x_affine_den = self.fp2_eval_horner(self.x_den, x, "x_den")
        x_affine = self.fp2_div(x_affine_num, x_affine_den)
        y_affine_num = self.fp2_eval_horner(self.y_num, x, "y_num")
        y_affine_den = self.fp2_eval_horner(self.y_den, x, "y_den")
        y_affine_eval = self.fp2_div(y_affine_num, y_affine_den)
        y_affine = self.fp2_mul(y_affine_eval, y)

        return x_affine, y_affine


class IsogenyG1(ModuloCircuit):
    def __init__(self, name: str, curve_id: int, compilation_mode: int = 0):
        super().__init__(
            name=name,
            curve_id=curve_id,
            compilation_mode=compilation_mode,
            generic_circuit=False,
        )
        self.curve = CURVES[curve_id]

    def set_consts(self):
        """
        Sets the constants for the isogeny to G1 map.
        """
        x_rational, y_rational = get_isogeny_to_g1_map(CurveID(self.curve_id))
        self.x_num = [
            self.set_or_get_constant(c) for c in x_rational.numerator.coefficients
        ]
        self.x_den = [
            self.set_or_get_constant(c) for c in x_rational.denominator.coefficients
        ]

        self.y_num = [
            self.set_or_get_constant(c) for c in y_rational.numerator.coefficients
        ]
        self.y_den = [
            self.set_or_get_constant(c) for c in y_rational.denominator.coefficients
        ]

    def run_isogeny(
        self, x: ModuloCircuitElement, y: ModuloCircuitElement
    ) -> tuple[ModuloCircuitElement, ModuloCircuitElement]:
        """
        Runs the isogeny to G1 map.
        """
        self.set_consts()

        x_affine_num = self.eval_horner(self.x_num, x, "x_num")
        x_affine_den = self.eval_horner(self.x_den, x, "x_den")
        x_affine = self.div(x_affine_num, x_affine_den)
        y_affine_num = self.eval_horner(self.y_num, x, "y_num")
        y_affine_den = self.eval_horner(self.y_den, x, "y_den")
        y_affine_eval = self.div(y_affine_num, y_affine_den)
        y_affine = self.mul(y_affine_eval, y)

        return x_affine, y_affine


if __name__ == "__main__":
    circuit = IsogenyG1("isogeny", 1)  # BLS12-381

    field_x = circuit.write_element(
        circuit.field(
            1412853964218444964438936699552956047210482383152224645596624291427056376487356261681298103080878386132407858666637
        )
    )
    field_y = circuit.write_element(
        circuit.field(
            752734926215712395741522221355891264404138695398702662135908094550118515106801651502315795564392519475687558113863
        )
    )

    x_affine, y_affine = circuit.run_isogeny(field_x, field_y)



def neg_3_base_le(scalar: int) -> list[int]:
    """
    Decomposes a scalar into base -3 representation.
    :param scalar: The integer to be decomposed.
    :return: A list of coefficients in base -3 representation. (Least significant bit first),
    with digits [-1, 0, 1] such that scalar = sum((-3) ** i * d for (i, d) in enumerate(digits))
    """
    if scalar == 0:
        return [0]
    digits = []
    while scalar != 0:
        remainder = scalar % 3
        if (
            remainder == 2
        ):  # if the remainder is 2, we set it to -1 and add 1 to the next digit
            remainder = -1
            scalar += 1
        # For remainder 1 and 0, no change is required
        digits.append(remainder)
        scalar = -(scalar // 3)  # divide by -3 for the next digit

    return digits


def construct_digit_vectors(es: list[int]) -> list[list[int]]:
    dss_ = [neg_3_base_le(e) for e in es]  # Base -3 digits
    max_len = max(len(ds) for ds in dss_)
    dss_ = [ds + [0] * (max_len - len(ds)) for ds in dss_]
    # Transposing the matrix
    dss = [[dss_[row][col] for row in range(len(dss_))] for col in range(max_len)]
    return dss


def positive_negative_multiplicities(digits: list[int]) -> tuple[int, int]:

    sum_p = sum((-3) ** i for (i, d) in enumerate(digits) if d == 1)
    sum_n = sum((-3) ** i for (i, d) in enumerate(digits) if d == -1)

    return (sum_p, sum_n)


def scalar_to_base_neg3_le(u128: int) -> tuple[int, int, int, int]:
    def sign(felt252: int) -> int:
        # In cairo, the felt252 :
        # - is considered positive if it is in [0, STARK//2[
        # - is considered negative if it is in ]STARK//2, STARK[
        # Where STARK = 2**251+ 17* 2**192 + 1
        # Note :  value being exactly STARK//2 will not happen in the construction here.
        if felt252 > 0:
            return 1
        elif felt252 < 0:
            return -1
        else:
            # Note : the case where the input is 0, the sign does not matter.
            # Any value could be returned, so choose whatever is more performant in implementation.
            return 0

    digits = neg_3_base_le(u128)

    # Even if input is u128, both sum_p and sum_n might be larger and negative. Output of positive_negative_multiplicities should be felt252.

    sum_p, sum_n = positive_negative_multiplicities(digits)

    # return type : felt252, felt252, felt252, felt252
    return (abs(sum_p), abs(sum_n), sign(sum_p), sign(sum_n))


if __name__ == "__main__":
    print(positive_negative_multiplicities(neg_3_base_le(2**128)))



from dataclasses import dataclass

from garaga.algebra import PyFelt
from garaga.definitions import CURVES, CurveID, G1Point, get_base_field
from garaga.hints.io import bigint_split, int_to_u384
from garaga.signature import apply_isogeny, hash_to_field


@dataclass(slots=True)
class MapToCurveHint:
    gx1_is_square: bool
    y1: PyFelt
    y_flag: bool

    def to_cairo(self) -> str:
        return f"MapToCurveHint {{ gx1_is_square: {str(self.gx1_is_square).lower()}, y1: {int_to_u384(self.y1.value, as_hex=True)}, y_flag: {str(self.y_flag).lower()} }}"

    def to_calldata(self) -> list[int]:
        cd = []
        cd.append(int(self.gx1_is_square))
        cd.extend(bigint_split(self.y1.value))
        cd.append(int(self.y_flag))
        return cd


@dataclass(slots=True)
class HashToCurveHint:
    f0_hint: MapToCurveHint
    f1_hint: MapToCurveHint

    def to_cairo(self) -> str:
        return f"""HashToCurveHint {{ f0_hint: {self.f0_hint.to_cairo()},
        f1_hint: {self.f1_hint.to_cairo()} }}"""

    def to_calldata(self) -> list[int]:
        cd = []
        cd.extend(self.f0_hint.to_calldata())
        cd.extend(self.f1_hint.to_calldata())
        return cd


def build_map_to_curve_hint(u: PyFelt) -> tuple[G1Point, MapToCurveHint]:
    field = get_base_field(CurveID.BLS12_381)
    a = field(CURVES[CurveID.BLS12_381.value].swu_params.A)
    b = field(CURVES[CurveID.BLS12_381.value].swu_params.B)
    z = field(CURVES[CurveID.BLS12_381.value].swu_params.Z)

    zeta_u2 = z * u**2
    ta = zeta_u2**2 + zeta_u2
    num_x1 = b * (ta + field.one())

    if ta.value == 0:
        div = a * z
    else:
        div = a * -ta

    num2_x1 = num_x1**2
    div2 = div**2
    div3 = div2 * div
    assert div3.value != 0

    num_gx1 = (num2_x1 + a * div2) * num_x1 + b * div3
    num_x2 = zeta_u2 * num_x1

    gx1 = num_gx1 / div3
    gx1_square = gx1.is_quad_residue()
    if gx1_square:
        y1 = gx1.sqrt(min_root=False)
        assert y1 * y1 == gx1
    else:
        y1 = (z * gx1).sqrt(min_root=False)
        assert y1 * y1 == z * gx1

    y2 = zeta_u2 * u * y1
    y = y1 if gx1_square else y2
    y_flag = y.value % 2 == u.value % 2

    num_x = num_x1 if gx1_square else num_x2
    x_affine = num_x / div
    y_affine = -y if y.value % 2 != u.value % 2 else y

    point_on_curve = G1Point(
        x_affine.value, y_affine.value, CurveID.BLS12_381, iso_point=True
    )
    return point_on_curve, MapToCurveHint(
        gx1_is_square=gx1_square, y1=y1, y_flag=y_flag
    )


def build_hash_to_curve_hint(message: bytes) -> HashToCurveHint:
    felt0, felt1 = hash_to_field(message, 2, CurveID.BLS12_381.value, "sha256")
    pt0, f0_hint = build_map_to_curve_hint(felt0)
    pt1, f1_hint = build_map_to_curve_hint(felt1)
    sum_pt = pt0.add(pt1)
    # print(
    #     f"sum_pt: {int_to_u384(sum_pt.x, as_hex=False)} {int_to_u384(sum_pt.y, as_hex=False)}"
    # )
    sum_pt = apply_isogeny(sum_pt)
    # print(
    #     f"sum_pt: {int_to_u384(sum_pt.x, as_hex=False)} {int_to_u384(sum_pt.y, as_hex=False)}"
    # )
    x = CURVES[CurveID.BLS12_381.value].x
    n = CURVES[CurveID.BLS12_381.value].n
    cofactor = (1 - (x % n)) % n  # 15132376222941642753
    # print(f"cofactor: {cofactor}, hex :{hex(cofactor)}")

    return HashToCurveHint(
        f0_hint=f0_hint,
        f1_hint=f1_hint,
    )


if __name__ == "__main__":
    field = get_base_field(CurveID.BLS12_381)

    import hashlib

    hint = build_hash_to_curve_hint(hashlib.sha256(b"Hello, World!").digest())
    print(hint.to_cairo())
    # print(hint.to_calldata())



import functools

from garaga.algebra import FunctionFelt, ModuloCircuitElement, PyFelt

PRIME = 2**251 + 17 * 2**192 + 1  # STARK prime


def assert_integer(val):
    """
    Asserts that the input is an integer (and not relocatable value).
    """
    assert isinstance(val, int), f"Expected integer, found: {val}."


def as_int(val, prime):
    """
    Returns the lift of the given field element, val, as an integer in the range
    (-prime/2, prime/2).
    """
    assert_integer(val)
    return val if val < prime // 2 else val - prime


def to_hex_str(value: str | int):
    if isinstance(value, str):
        value = value.strip()  # Trim whitespaces
        if value.lower().startswith("0x"):
            try:
                # Validate hexadecimal and return in lowercase
                return "0x" + hex(int(value, 16))[2:].lower()
            except ValueError:
                raise ValueError(f"Invalid hexadecimal value: {value}")
        else:
            try:
                # Convert decimal string to int then to hex
                return hex(int(value)).lower()
            except ValueError:
                raise ValueError(f"Invalid decimal value: {value}")
    elif isinstance(value, int):
        return hex(value).lower()
    else:
        raise TypeError(f"Expected str or int, got {type(value).__name__}")


# Split a bigint into its limbs.
# Accepts int, ModuloCircuitElement, PyFelt, or bytes.
# Returns the limbs in little-endian order.
# Ie : x = d0 + d1 * base + d2 * base^2 + ... + dn * base^n
# Returns the coefficients [d0, d1, d2, ..., dn]
def bigint_split(
    x: int | ModuloCircuitElement | PyFelt | bytes, n_limbs: int = 4, base: int = 2**96
) -> list[int]:
    if isinstance(x, int):
        pass
    elif isinstance(x, (ModuloCircuitElement, PyFelt)):
        x = x.value
    elif isinstance(x, bytes):
        x = int.from_bytes(x, byteorder="big")
    else:
        raise ValueError(f"Invalid type for bigint_split: {type(x)}")

    coeffs = []
    degree = n_limbs - 1
    for n in range(degree, 0, -1):
        q, r = divmod(x, base**n)
        coeffs.append(q)
        x = r
    coeffs.append(x)
    return coeffs[::-1]


def bytes_to_u32_array(bytes_array: bytes, name: str) -> str:
    code = f"const {name}: [u32; {((len(bytes_array) + 3) // 4)}] = ["

    for i in range(0, len(bytes_array), 4):
        chunk = bytes_array[i : i + 4]

        u32_value = int.from_bytes(chunk, "big")
        code += f"{hex(u32_value)}, "

    code += "];"
    return code


def to_int(value: str | int | bytes) -> int:
    """
    Convert a string or integer to an integer. Supports hexadecimal and decimal strings.
    """
    if isinstance(value, str):
        value = value.strip()  # Trim whitespaces
        if value.lower().startswith("0x"):
            try:
                return int(value, 16)
            except ValueError:
                raise ValueError(f"Invalid hexadecimal value: {value}")
        else:
            try:
                return int(value)
            except ValueError:
                raise ValueError(f"Invalid decimal value: {value}")
    elif isinstance(value, int):
        return value
    elif isinstance(value, bytes):
        return int.from_bytes(value, byteorder="big")
    else:
        raise TypeError(f"Expected str, int, or bytes, got {type(value).__name__}")


def int_to_u2XX(x: int | PyFelt, curve_id: int = 0, as_hex=True) -> str:
    if curve_id == 1:
        return int_to_u384(x, as_hex)
    else:
        return int_to_u288(x, as_hex)


def int_to_u288(x: int | PyFelt, as_hex=True) -> str:
    limbs = bigint_split(x, 3, 2**96)
    if as_hex:
        return f"u288{{limb0:{hex(limbs[0])}, limb1:{hex(limbs[1])}, limb2:{hex(limbs[2])}}}"
    else:
        return f"u288{{limb0:{limbs[0]}, limb1:{limbs[1]}, limb2:{limbs[2]}}}"


def int_to_u384(x: int | PyFelt, as_hex=True) -> str:
    limbs = bigint_split(x, 4, 2**96)
    if as_hex:
        return f"u384{{limb0:{hex(limbs[0])}, limb1:{hex(limbs[1])}, limb2:{hex(limbs[2])}, limb3:{hex(limbs[3])}}}"
    else:
        return f"u384{{limb0:{limbs[0]}, limb1:{limbs[1]}, limb2:{limbs[2]}, limb3:{limbs[3]}}}"


def int_to_u256(x: int | PyFelt) -> str:
    assert 0 <= x < 2**256, f"Value {x} is too large to fit in a u256"
    limbs = bigint_split(x, 2, 2**128)
    return f"u256{{low:{hex(limbs[0])}, high:{hex(limbs[1])}}}"


def int_to_u128(x: int | PyFelt) -> str:
    assert 0 <= x < 2**128, f"Value {x} is too large to fit in a u128"
    return hex(x)


def int_array_to_u256_array(x: list[int] | list[PyFelt]) -> str:
    return f"array![{', '.join([int_to_u256(i) for i in x])}]"


def int_array_to_u128_array(x: list[int] | list[PyFelt]) -> str:
    return f"array![{', '.join([int_to_u128(i) for i in x])}]"


def int_array_to_u384_array(x: list[int] | list[PyFelt], const=False) -> str:
    if const:
        return f"[{', '.join([int_to_u384(i) for i in x])}]"
    else:
        return f"array![{', '.join([int_to_u384(i) for i in x])}]"


def int_array_to_u288_array(x: list[int] | list[PyFelt], const=False) -> str:
    if const:
        return f"[{', '.join([int_to_u288(i) for i in x])}]"
    else:
        return f"array![{', '.join([int_to_u288(i) for i in x])}]"


def int_array_to_u2XX_array(
    x: list[int] | list[PyFelt], curve_id: int, const=False
) -> str:
    if curve_id == 1:
        return int_array_to_u384_array(x, const)
    else:
        return int_array_to_u288_array(x, const)


def bigint_pack(x: object, n_limbs: int, base: int) -> int:
    val = 0
    for i in range(n_limbs):
        val += as_int(getattr(x, f"d{i}"), PRIME) * base**i
    return val


def bigint_pack_ptr(memory: object, ptr: object, n_limbs: int, base: int) -> int:
    val = 0
    for i in range(n_limbs):
        val += as_int(memory[ptr + i], PRIME) * base**i
    return val


def pack_e12d(x: object, n_limbs: int, base: int):
    val = []
    for i in range(12):
        val.append(bigint_pack(getattr(x, f"w{i}"), n_limbs, base))
    return val


def pack_e6d(x: object, n_limbs, base):
    val = []
    for i in range(6):
        val.append(bigint_pack(getattr(x, f"v{i}"), n_limbs, base))
    return val


# ptr : Uint384*
def pack_bigint_array(
    ptr: object, n_limbs: int, base: int, n_elements: int, offset: int = 0
):
    val = []
    for i in range(n_elements):
        val.append(bigint_pack(ptr[offset + i], n_limbs, base))
    return val


def pack_felt_ptr(memory: object, ptr: object, n_elements: int):
    val = []
    for i in range(n_elements):
        val.append(memory[ptr + i])
    return val


def pack_bigint_ptr(
    memory: object,
    ptr: object,
    n_limbs: int,
    base: int,
    n_elements: int,
):
    val = []
    for i in range(n_elements):
        val.append(bigint_pack_ptr(memory, ptr + i * n_limbs, n_limbs, base))
    return val


#### WRITE HINTS


def bigint_fill(x: int, ids: object, n_limbs: int, base: int):
    xs = bigint_split(x, n_limbs, base)
    for i in range(n_limbs):
        setattr(ids, f"d{i}", xs[i])
    return


def fill_felt_ptr(x: list, memory: object, address: int):
    for i in range(len(x)):
        memory[address + i] = x[i]


def fill_limbs(limbs: list, ids: object):
    """
    limbs: list of integers
    ids: reference to Cairo object
    """
    for i, limb in enumerate(limbs):
        setattr(ids, f"d{i}", limb)
    return


def bigint_split_array(
    x: list[int | PyFelt | ModuloCircuitElement],
    n_limbs: int = 4,
    base: int = 2**96,
    prepend_length=False,
) -> list[int]:
    xs = []
    if prepend_length:
        xs.append(len(x))
    for e in x:
        xs.extend(bigint_split(e, n_limbs, base))
    return xs


def fill_bigint_array_into_felt_ptr(
    x: list, memory: object, address: int, base: int, n_limbs: int
):
    felts = []
    for val in x:
        felts.extend(bigint_split(val, n_limbs, base))
    fill_felt_ptr(felts, memory, address)


def fill_e6d(x: list, ids: object, n_limbs: int, base: int):
    for i in range(6):
        bigint_fill(x[i], getattr(ids, f"v{i}"), n_limbs, base)
    return


def fill_e12d(x: list, ids: object, n_limbs: int, base: int):
    for i in range(12):
        bigint_fill(x[i], getattr(ids, f"w{i}"), n_limbs, base)
    return


def fill_uint256(x: int, ids: object):
    x0, x1 = split_128(x)
    setattr(ids, "low", x0)
    setattr(ids, "high", x1)
    return


def padd_function_felt(
    f: FunctionFelt, n: int, py_felt: bool = False, batched: bool = False
) -> tuple[list[int], list[int], list[int], list[int]]:
    a_num = f.a.numerator.get_coeffs() if py_felt else f.a.numerator.get_value_coeffs()
    a_den = (
        f.a.denominator.get_coeffs() if py_felt else f.a.denominator.get_value_coeffs()
    )
    b_num = f.b.numerator.get_coeffs() if py_felt else f.b.numerator.get_value_coeffs()
    b_den = (
        f.b.denominator.get_coeffs() if py_felt else f.b.denominator.get_value_coeffs()
    )
    assert len(a_num) <= n + 1 + (
        2 if batched else 0
    ), f"a_num has {len(a_num)} limbs, expected at most {n + 1 + (2 if batched else 0)}"
    assert len(a_den) <= n + 2 + (
        2 if batched else 0
    ), f"a_den has {len(a_den)} limbs, expected at most {n + 2 + (2 if batched else 0)}"
    assert len(b_num) <= n + 2 + (
        2 if batched else 0
    ), f"b_num has {len(b_num)} limbs, expected at most {n + 2 + (2 if batched else 0)}"
    assert len(b_den) <= n + 5 + (
        2 if batched else 0
    ), f"b_den has {len(b_den)} limbs, expected at most {n + 5 + (2 if batched else 0)}"
    zero = [f.a.numerator.field.zero()] if py_felt else [0]
    a_num = a_num + zero * (n + 1 + (2 if batched else 0) - len(a_num))
    a_den = a_den + zero * (n + 2 + (2 if batched else 0) - len(a_den))
    b_num = b_num + zero * (n + 2 + (2 if batched else 0) - len(b_num))
    b_den = b_den + zero * (n + 5 + (2 if batched else 0) - len(b_den))
    return (a_num, a_den, b_num, b_den)


def fill_sum_dlog_div(f: FunctionFelt, n: int, ref: object, segments: object):
    a_num, a_den, b_num, b_den = padd_function_felt(f, n)

    ref.a_num = segments.gen_arg(bigint_split_array(a_num, 4, 2**96))
    ref.a_den = segments.gen_arg(bigint_split_array(a_den, 4, 2**96))
    ref.b_num = segments.gen_arg(bigint_split_array(b_num, 4, 2**96))
    ref.b_den = segments.gen_arg(bigint_split_array(b_den, 4, 2**96))


def fill_g1_point(p: tuple[int, int], ref: object):
    bigint_fill(p[0], ref.x, 4, 2**96)
    bigint_fill(p[1], ref.y, 4, 2**96)
    return


### OTHERS


def flatten(t):
    result = []
    for item in t:
        if isinstance(item, (tuple, list)):
            result.extend(flatten(item))
        else:
            result.append(item)
    return result


def split_128(a: int) -> tuple[int, int]:
    assert (
        0 <= a < 2**256
    ), f"Value {hex(a)} {a.bit_length()}bits is too large to fit in a u256"
    """Takes in value, returns uint256-ish tuple (low, high)."""
    return (a & ((1 << 128) - 1), a >> 128)


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)



import re
import subprocess
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor
from typing import Type

from garaga.definitions import CurveID, get_base_field
from garaga.hints.io import int_array_to_u384_array
from garaga.modulo_circuit import ModuloCircuit, ModuloCircuitElement, PyFelt
from garaga.modulo_circuit_structs import Cairo1SerializableStruct


def get_circuit_definition_impl_template(num_outputs: int):
    TEMPLATE = """\n
    impl CircuitDefinition{num_outputs}<
        {elements}
    > of core::circuit::CircuitDefinition<
        (
            {ce_elements}
        )
    > {{
        type CircuitType =
            core::circuit::Circuit<
                ({elements_tuple},)
            >;
    }}
    impl MyDrp_{num_outputs}<
        {elements}
    > of Drop<
        (
            {ce_elements}
        )
    >;
    """
    elements = ", ".join(f"E{i}" for i in range(num_outputs))
    ce_elements = ", ".join(f"CE<E{i}>" for i in range(num_outputs))
    elements_tuple = ", ".join(f"E{i}" for i in range(num_outputs))
    return TEMPLATE.format(
        num_outputs=num_outputs,
        elements=elements,
        ce_elements=ce_elements,
        elements_tuple=elements_tuple,
    )


class BaseModuloCircuit(ABC):
    """
    Base class for all modulo circuits that will be compiled to Cairo code.
    Parameters:
    - name: str, the name of the circuit
    - curve_id: int, the id of the curve
    - auto_run: bool, whether to run the circuit automatically at initialization.
                When compiling, this flag is set to true so the ModuloCircuit class inside the
                 ".circuit" member of this class holds the necessary metadata
                about the operations that will be compiled.
                For CairoZero, this flag will be set to False in the Python hint, so that
                BaseModuloCircuit.run_circuit() can be called on a segment parsed from the
                CairoZero VM.
    - compilation mode: 0 (CairoZero) or 1 (Cairo)
    """

    def __init__(
        self,
        name: str,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 0,
    ) -> None:
        self.name = name
        self.curve_id = curve_id
        self.field = get_base_field(curve_id)
        self.init_hash = None
        self.generic_over_curve = False
        self.compilation_mode = compilation_mode
        if auto_run:
            self.input = self.build_input()
            self.circuit: ModuloCircuit = self._run_circuit_inner(self.input.copy())

    @abstractmethod
    def build_input(self) -> list[PyFelt]:
        """
        This method is used to create the necessary inputs that will be written to the ModuloCircuit.
        It works in pair with the _run_circuit_inner function, where the _run_circuit_inner will use the output of
        the build_input function to "deserialize" the list of elements and write them to the ModuloCircuit class.
        """

    @property
    def full_input_cairo1(self) -> list[PyFelt] | list[Cairo1SerializableStruct]:
        if self.circuit is None:
            raise ValueError("Circuit not run yet")
        if self.circuit.input_structs:
            return self.circuit.input_structs
        else:
            res = self.input + self.circuit.witnesses
        return res

    @abstractmethod
    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        """
        This method is responsible for
        - deserializing the input list of elements,
        - creating a ModuloCircuit class (or class that derives from ModuloCircuit)
        - "writing" the inputs to the ModuloCircuit class to obtain ModuloCircuitElements
        - using the methods add, sub, mul, inv (or higher level methods) of the ModuloCircuit class
            to define the list of operations on the given inputs
        - Returning the ModuloCircuit class in a state where the circuit has been run, and therefore holding
            the metadata so that its instructions can be compiled to Cairo code.

        """

    def run_circuit(self, input: list[int]) -> ModuloCircuit:
        """
        A simple wrapper around _run_circuit_inner that converts the input as a list of ints to a list of PyFelt.
        Used in CairoZero hint.
        """
        # print(
        #     f"Running circuit for {self.name} with CurveID {CurveID(self.curve_id).name}..."
        # )
        circuit_input = [self.field(x) for x in input]
        return self._run_circuit_inner(circuit_input)


class BaseEXTFCircuit(BaseModuloCircuit):
    """
    A extension of the BaseModuloCircuit class that holds an init_hash, used for CairoZero.
    Not relevant for Cairo1 circuits.
    """

    def __init__(
        self,
        name: str,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(name, curve_id, auto_run, compilation_mode)
        self.init_hash = init_hash


def compilation_mode_to_file_header(mode: int, curve_ids: set[CurveID]) -> str:
    if mode == 0:
        return """
from starkware.cairo.common.registers import get_fp_and_pc, get_label_location
from modulo_circuit import ExtensionFieldModuloCircuit, ModuloCircuit, get_void_modulo_circuit, get_void_extension_field_modulo_circuit
from definitions import bn, bls
"""
    elif mode == 1:
        moduluses = [
            f"get_{curve_id.name}_modulus"
            for curve_id in sorted(curve_ids, key=lambda x: x.name)
        ]
        return f"""
use core::circuit::{{
    RangeCheck96, AddMod, MulMod, u384, u96, circuit_add, circuit_sub,
    circuit_mul, circuit_inverse, EvalCircuitResult, EvalCircuitTrait, CircuitOutputsTrait,
    CircuitModulus, AddInputResultTrait, CircuitInputs, CircuitDefinition,
    CircuitData, CircuitInputAccumulator
}};
use garaga::core::circuit::{{AddInputResultTrait2, u288IntoCircuitInputValue, IntoCircuitInputValue}};
use core::circuit::CircuitElement as CE;
use core::circuit::CircuitInput as CI;
use garaga::definitions::{{get_a, get_b, get_modulus, get_g, get_min_one, G1Point, G2Point, E12D, u288, E12DMulQuotient, G1G2Pair, BNProcessedPair, BLSProcessedPair, MillerLoopResultScalingFactor, G2Line, E12T, {', '.join(moduluses)}}};
use core::option::Option;
"""


def cairo1_tests_header() -> str:
    return """
#[cfg(test)]
mod tests {
    use core::traits::TryInto;

    use core::circuit::{
        RangeCheck96, AddMod, MulMod, u96, CircuitElement, CircuitInput, circuit_add, circuit_sub,
        circuit_mul, circuit_inverse, EvalCircuitResult, EvalCircuitTrait, u384,
        CircuitOutputsTrait, CircuitModulus, AddInputResultTrait, CircuitInputs
    };
    use garaga::definitions::{G1Point, G2Point, E12D, E12DMulQuotient, G1G2Pair, BNProcessedPair, BLSProcessedPair, MillerLoopResultScalingFactor, G2Line};

"""


def to_snake_case(s: str) -> str:
    return re.sub(r"(?<=[a-z])(?=[A-Z])|[^a-zA-Z0-9]", "_", s).lower()


def create_circuit_instances(
    circuit_class: Type[BaseModuloCircuit],
    curve_id: CurveID,
    params: list[dict],
    compilation_mode: int,
) -> list[BaseModuloCircuit]:
    """
    Create a list of circuit instances from a given circuit class, curve id, params and compilation mode.
    """
    circuits = []
    if params is None:
        circuit_instance = circuit_class(
            curve_id=curve_id.value, compilation_mode=compilation_mode, auto_run=True
        )
        circuits.append(circuit_instance)
    else:
        for param in params:
            circuit_instance = circuit_class(
                curve_id=curve_id.value,
                compilation_mode=compilation_mode,
                auto_run=True,
                **param,
            )
            circuits.append(circuit_instance)
    return circuits


def compile_single_circuit(
    circuit_instance: BaseModuloCircuit,
) -> tuple[ModuloCircuit, str]:
    """
    Compile a single circuit instance to Cairo code.
    Returns the compiled circuit and the full function name.
    """
    curve_id = CurveID(circuit_instance.curve_id)
    function_name = (
        f"{circuit_instance.name.upper()}"
        if circuit_instance.circuit.generic_circuit
        else f"{curve_id.name}_{circuit_instance.name.upper()}"
    )
    compiled_circuit, full_function_name = circuit_instance.circuit.compile_circuit(
        function_name=function_name
    )
    return compiled_circuit, full_function_name


def compile_circuit(
    curve_id: CurveID,
    circuit_class: BaseModuloCircuit,
    params: list[dict],
    compilation_mode: int,
) -> tuple[list[str], list[str], list[BaseModuloCircuit]]:
    """
    Compile a list of circuit instances to Cairo code.
    Returns :
    - compiled_circuits: list of compiled circuits as strings
    - full_function_names: list of full function names as strings
    - circuit_instances: list of circuit instances that have been compiled
    """
    circuits = create_circuit_instances(
        circuit_class, curve_id, params, compilation_mode
    )
    compiled_circuits = []
    full_function_names = []

    for circuit_instance in circuits:
        compiled_circuit, full_function_name = compile_single_circuit(circuit_instance)
        compiled_circuits.append(compiled_circuit)
        full_function_names.append(full_function_name)

    return compiled_circuits, full_function_names, circuits


def create_cairo1_test(function_name: str, input: list, output: list, curve_id: int):
    return ""
    if function_name == "":
        # print(f"passing test")
        return ""
    include_curve_id = CurveID.find_value_in_string(function_name) == None
    if all(isinstance(i, Cairo1SerializableStruct) for i in input):
        input_code = "\n".join([i.serialize() for i in input])
        input_params = ",".join([f"{i.name}" for i in input])
    elif all(isinstance(i, PyFelt) for i in input):
        input_code = f"let input = {int_array_to_u384_array(input)};"
        input_params = "input"
    else:
        raise ValueError(
            f"Invalid input type for Cairo1 test : {[type(i) for i in input]}"
        )
    if all(isinstance(i, Cairo1SerializableStruct) for i in output):
        output_code = f"let ({', '.join([i.name+'_result' for i in output])}) = {function_name}({input_params}{', ' + str(curve_id) if include_curve_id else ''});"
        output_code += "\n".join([i.serialize() for i in output])
        for struct in output:
            output_code += f"assert_eq!({struct.name}_result, {struct.name});"
    elif all(isinstance(i, ModuloCircuitElement) for i in output):
        output_code = f"let got = {function_name}({input_params}{', ' + str(curve_id) if include_curve_id else ''});"
        output_code += f"let exp = {int_array_to_u384_array(output)};"
        output_code += "assert_eq!(got.len(), exp.len());"
        output_code += "assert_eq!(got, exp);"
    else:
        raise ValueError(
            f"Invalid output type for Cairo1 test : {[type(i) for i in output]}"
        )
    code = f"""
    #[test]
    fn test_{function_name}_{CurveID(curve_id).name}() {{
        {input_code}
        {output_code}
    }}
    """

    return code


def format_cairo_files_in_parallel(
    filenames, compilation_mode, precompiled_circuits_dir
):
    if compilation_mode == 0:
        print("Formatting .cairo zero files in parallel...")
        cairo_files = [f"{precompiled_circuits_dir}{f}.cairo" for f in filenames]
        with ProcessPoolExecutor() as executor:
            futures = [
                executor.submit(
                    subprocess.run, ["cairo-format", file, "-i"], check=True
                )
                for file in cairo_files
            ]
            for future in futures:
                future.result()  # Wait for all formatting tasks to complete
        print("Done!")
    elif compilation_mode == 1:
        subprocess.run(["scarb", "fmt", f"{precompiled_circuits_dir}"], check=True)



from dataclasses import dataclass, field
from enum import Enum

from garaga.algebra import Polynomial, PyFelt
from garaga.definitions import N_LIMBS, get_irreducible_poly
from garaga.hints.extf_mul import (
    nondeterministic_extension_field_div,
    nondeterministic_extension_field_mul_divmod,
)
from garaga.modulo_circuit import BATCH_SIZE, ModuloCircuitElement, WriteOps
from garaga.poseidon_transcript import CairoPoseidonTranscript
from garaga.precompiled_circuits.fp2 import Fp2Circuits

POSEIDON_BUILTIN_SIZE = 6
POSEIDON_OUTPUT_S1_INDEX = 4


# Represents the state of the accumulation of the equation
#  c_i * Π(Pi(z)) = c_i*Q_i*P + c_i*R_i inside the circuit.
# Only store ci*Π(Pi(z)) (as Emulated Field Element) and ci*R_i (as Polynomial)
@dataclass(slots=True)
class EuclideanPolyAccumulator:
    lhs: ModuloCircuitElement
    R: list[ModuloCircuitElement]
    R_evaluated: ModuloCircuitElement


class AccPolyInstructionType(Enum):
    MUL = "MUL"
    DIV = "DIV"
    SQUARE_TORUS = "SQUARE_TORUS"


@dataclass(slots=True)
class AccumulatePolyInstructions:
    Pis: list[list[list[ModuloCircuitElement]]] = field(default_factory=list)
    Qis: list[Polynomial] = field(default_factory=list)
    Ris: list[list[ModuloCircuitElement]] = field(default_factory=list)
    Ps_sparsities: list[None | list[list[int]]] = field(default_factory=list)
    r_sparsities: list[None | list[int]] = field(default_factory=list)
    types: list[AccPolyInstructionType] = field(default_factory=list)
    n: int = field(default=0)
    rlc_coeffs: list[ModuloCircuitElement] = field(default_factory=list)
    Pis_of_Z: list[ModuloCircuitElement] = field(default_factory=list)

    def append(
        self,
        type: AccPolyInstructionType,
        Pis: list[list[ModuloCircuitElement]],
        Q: Polynomial,
        R: list[ModuloCircuitElement],
        Ps_sparsities: None | list[list[int] | None] = None,
        r_sparsity: None | list[int] = None,
    ):
        if type != AccPolyInstructionType.MUL:
            assert len(Pis) == 2
        self.types.append(type)
        self.Pis.append(Pis)
        self.Qis.append(Q)
        self.Ris.append(R)
        self.Ps_sparsities.append(Ps_sparsities)
        self.r_sparsities.append(r_sparsity)
        self.n += 1


class ExtensionFieldModuloCircuit(Fp2Circuits):
    def __init__(
        self,
        name: str,
        curve_id: int,
        extension_degree: int,
        init_hash: int = None,
        hash_input: bool = True,
        compilation_mode: int = 0,
        generic_circuit: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            curve_id=curve_id,
            compilation_mode=compilation_mode,
            generic_circuit=generic_circuit,
        )
        self.class_name = "ExtensionFieldModuloCircuit"
        self.extension_degree = extension_degree
        self.z_powers: list[ModuloCircuitElement] = []
        self.acc: list[EuclideanPolyAccumulator] = [
            self._init_accumulator(self.extension_degree),
            self._init_accumulator(self.extension_degree * 2),
        ]
        self.hash_input = hash_input
        self.transcript: CairoPoseidonTranscript = CairoPoseidonTranscript(
            init_hash=(
                int.from_bytes(self.name.encode(), "big")
                if init_hash is None
                else init_hash
            )
        )
        self.z_powers: list[ModuloCircuitElement] = []
        self.ops_counter = {
            "Circuit": self.name,
            "EXTF_SQUARE": 0,
            "EXTF_MUL_DENSE": 0,
            # "EXTF_DIV": 0,
        }
        self.accumulate_poly_instructions: list[AccumulatePolyInstructions] = [
            AccumulatePolyInstructions(),
            AccumulatePolyInstructions(),
        ]

    def _init_accumulator(self, extension_degree: int = None):
        extension_degree = extension_degree or self.extension_degree
        # Todo : Add compilation mode 1 support
        if self.compilation_mode == 1:
            return EuclideanPolyAccumulator(
                lhs=None,
                R=[None] * extension_degree,
                R_evaluated=None,
            )
        else:
            return EuclideanPolyAccumulator(
                lhs=self.set_or_get_constant(0),
                R=[None] * extension_degree,
                R_evaluated=self.set_or_get_constant(0),
            )

    @property
    def commitments(self):
        return [
            self.values_segment.segment_stacks[WriteOps.COMMIT][offset].felt
            for offset in sorted(self.values_segment.segment_stacks[WriteOps.COMMIT])
        ]

    @property
    def circuit_input(self):
        return [
            self.values_segment.segment_stacks[WriteOps.INPUT][offset].felt
            for offset in sorted(self.values_segment.segment_stacks[WriteOps.INPUT])
        ]

    def create_lines_z_powers(self, z: PyFelt, add_extf_power: bool = False):
        """
        Create powers of z for the evaluation of lines functions and the irreducible polynomial if add_extf_power is True.
        """
        powers = [z]  # z^1 at index 0
        if self.curve_id == 0:
            powers.append(self.square(z, "compute z^2"))  # z^2 at index 1
            powers.append(self.mul(powers[-1], z, "compute z^3"))  # z^3 at index 2
            powers.append(None)  # No z^4 at index 3
            powers.append(None)  # No z^5 at index 4
            powers.append(self.square(powers[2], "compute z^6"))  # z^6 at index 5
            powers.append(self.mul(powers[5], z, "compute z^7"))  # z^7 at index 6
            powers.append(None)  # No z^8 at index 7
            powers.append(
                self.mul(powers[6], powers[1], "compute z^9")
            )  # z^9 at index 8
            if add_extf_power:
                # Need to add z^12 :
                # z^12 = z^9 * z^3
                powers.append(None)  # No z^10 at index 9
                powers.append(None)  # No z^11 at index 10
                powers.append(self.mul(powers[8], powers[2], "compute z^12"))
            self.z_powers = powers

        elif self.curve_id == 1:
            # Need z^2, z^3, z^6, Z^8:
            powers.append(self.square(z, "compute z^2"))  # z^2 at index 1
            powers.append(self.mul(powers[-1], z, "compute z^3"))  # z^3 at index 2
            powers.append(None)  # No z^4 at index 3
            powers.append(None)  # No z^5 at index 4
            powers.append(self.square(powers[2], "compute z^6"))  # z^6 at index 5
            powers.append(None)  # No z^7 at index 6
            powers.append(
                self.mul(powers[5], powers[1], "compute z^8")
            )  # z^8 at index 7
            self.z_powers = powers
            if add_extf_power:
                # Need to add z^12 :
                # z^12 = z^7 * z^5
                powers.append(None)  # No z^9 at index 8
                powers.append(None)  # No z^10 at index 9
                powers.append(None)  # No z^11 at index 10
                powers.append(self.mul(powers[7], powers[4], "compute z^12"))
        else:
            raise ValueError(f"Invalid curve id: {self.curve_id}")

    def create_powers_of_Z(
        self,
        Z: PyFelt | ModuloCircuitElement,
        max_degree: int = None,
    ) -> list[ModuloCircuitElement]:
        if max_degree is None:
            max_degree = self.extension_degree
        if isinstance(Z, PyFelt):
            Z = self.write_cairo_native_felt(Z)
        elif isinstance(Z, int):
            Z = self.write_cairo_native_felt(self.field(Z))
        elif isinstance(Z, ModuloCircuitElement):
            pass
        else:
            raise ValueError(f"Invalid type for Z: {type(Z)}")
        powers = [Z]

        for i in range(2, max_degree + 1):
            powers.append(self.mul(powers[-1], powers[0], comment=f"Compute z^{i}"))

        self.z_powers = powers
        return powers

    def eval_poly_in_precomputed_Z(
        self,
        X: list[ModuloCircuitElement],
        sparsity: list[int] = None,
        poly_name: str = None,
    ) -> ModuloCircuitElement:
        """
        Evaluates a polynomial with coefficients `X` at precomputed powers of z.
        X(z) = X_0 + X_1 * z + X_2 * z^2 + ... + X_n * z^n

        If a `sparsity` list is provided, it is used to optimize the evaluation by skipping
        zero coefficients and applying special handling based on sparsity values:
        - A sparsity value of 0 indicates that the corresponding coefficient X_i is zero and should be skipped.
        - A sparsity value of 1 indicates that the corresponding coefficient X_i is non-zero and should be included in the evaluation.
        - A sparsity value of 2 indicates a special case where the term is directly a power of z, implying the coefficient X_i is 1 in this case.

        Parameters:
        - X: list[ModuloCircuitElement] - The coefficients of the polynomial to be evaluated.
        - sparsity: list[int] (optional) - A list indicating the sparsity of the polynomial. If None, the polynomial is treated as fully dense.

        Returns:
        - ModuloCircuitElement: The result of evaluating the polynomial at the precomputed powers of z.
        """
        if poly_name is None:
            poly_name = "UnnamedPoly"

        if sparsity:
            first_non_zero_idx = next(
                (i for i, x in enumerate(sparsity) if x != 0), None
            )
            if first_non_zero_idx == 0:
                match sparsity[0]:
                    case 0:
                        X_of_z = self.set_or_get_constant(0)
                    case 1:
                        X_of_z = X[0]
                    case 2:
                        X_of_z = self.set_or_get_constant(1)
                    case _:
                        raise ValueError(f"Invalid sparsity value: {sparsity[0]}")
            else:
                match sparsity[first_non_zero_idx]:
                    case 0:
                        X_of_z = self.set_or_get_constant(0)
                    case 1:
                        X_of_z = self.mul(
                            X[first_non_zero_idx],
                            self.z_powers[first_non_zero_idx - 1],
                            comment=f"Eval sparse poly {poly_name} step coeff_{first_non_zero_idx} * z^{first_non_zero_idx}",
                        )
                    case 2:
                        X_of_z = self.z_powers[first_non_zero_idx - 1]
                    case _:
                        raise ValueError(f"Invalid sparsity value: {sparsity[0]}")

            for i in range(first_non_zero_idx + 1, len(X)):
                match sparsity[i]:
                    case 1:
                        term = self.mul(
                            X[i],
                            self.z_powers[i - 1],
                            comment=f"Eval sparse poly {poly_name} step coeff_{i} * z^{i}",
                        )
                        add_comment = (
                            f"Eval sparse poly {poly_name} step + coeff_{i} * z^{i}"
                        )
                    case 2:
                        term = self.z_powers[
                            i - 1
                        ]  # In this case, sparsity[i] == 2 => X[i] = 1
                        add_comment = f"Eval sparse poly {poly_name} step + 1*z^{i}"
                    case 0:
                        continue
                    case _:
                        raise ValueError(f"Invalid sparsity value: {sparsity[i]}")
                X_of_z = self.add(
                    X_of_z,
                    term,
                    comment=add_comment,
                )
        else:
            X_of_z = self.eval_poly(X, self.z_powers, poly_name, "z")

        return X_of_z

    def extf_mul(
        self,
        Ps: list[list[ModuloCircuitElement]],
        extension_degree: int,
        Ps_sparsities: list[list[int] | None] = None,
        r_sparsity: list[int] = None,
        acc_index: int = 0,
    ) -> list[ModuloCircuitElement]:
        """
        Multiply in the extension field X * Y mod irreducible_poly
        Commit to R and add an EvalPolyInstruction to the accumulator.
        """
        assert (
            extension_degree > 2
        ), f"extension_degree={extension_degree} <= 2. Use self.mul or self.fp2_square instead."

        if Ps_sparsities is None:
            Ps_sparsities = [None] * len(Ps)
        assert len(Ps_sparsities) == len(
            Ps
        ), f"len(Ps_sparsities)={len(Ps_sparsities)} != len(Ps)={len(Ps)}"

        Q, R = nondeterministic_extension_field_mul_divmod(
            Ps, self.curve_id, extension_degree
        )

        R = self.write_elements(R, WriteOps.COMMIT, r_sparsity)

        if not any(sparsity for sparsity in Ps_sparsities) or not r_sparsity:
            self.ops_counter["EXTF_MUL_DENSE"] += 1

        self.accumulate_poly_instructions[acc_index].append(
            AccPolyInstructionType.MUL,
            Ps,
            Polynomial(Q),
            R,
            Ps_sparsities,
            r_sparsity,
        )
        return R

    def extf_div(
        self,
        X: list[ModuloCircuitElement],
        Y: list[ModuloCircuitElement],
        extension_degree: int,
        acc_index: int = 0,
    ) -> list[ModuloCircuitElement]:
        x_over_y = nondeterministic_extension_field_div(
            X, Y, self.curve_id, extension_degree
        )
        x_over_y = self.write_elements(x_over_y, WriteOps.COMMIT)

        Q, _ = nondeterministic_extension_field_mul_divmod(
            [x_over_y, Y], self.curve_id, extension_degree
        )
        # R should be X
        Q = Polynomial(Q)
        self.accumulate_poly_instructions[acc_index].append(
            AccPolyInstructionType.DIV, Pis=[x_over_y, Y], Q=Q, R=X
        )
        return x_over_y

    def extf_inv(
        self,
        Y: list[ModuloCircuitElement],
        extension_degree: int,
        acc_index: int = 0,
    ) -> list[ModuloCircuitElement]:
        one = [ModuloCircuitElement(self.field(1), -1)] + [
            ModuloCircuitElement(self.field(0), -1)
        ] * (extension_degree - 1)
        y_inv = nondeterministic_extension_field_div(
            one,
            Y,
            self.curve_id,
            extension_degree,
        )
        y_inv = self.write_elements(y_inv, WriteOps.COMMIT)

        Q, _ = nondeterministic_extension_field_mul_divmod(
            [y_inv, Y], self.curve_id, extension_degree
        )
        # R should be One. Passed at mocked modulo circuits element since fully determined by its sparsity.
        Q = Polynomial(Q)
        self.accumulate_poly_instructions[acc_index].append(
            AccPolyInstructionType.DIV,
            Pis=[y_inv, Y],
            Q=Q,
            R=one,
            r_sparsity=[2] + [0] * (extension_degree - 1),
        )
        return y_inv

    def conjugate_e12d(
        self, e12d: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        assert len(e12d) == 12
        return [
            e12d[0],
            self.neg(e12d[1]),
            e12d[2],
            self.neg(e12d[3]),
            e12d[4],
            self.neg(e12d[5]),
            e12d[6],
            self.neg(e12d[7]),
            e12d[8],
            self.neg(e12d[9]),
            e12d[10],
            self.neg(e12d[11]),
        ]

    def update_LHS_state(
        self,
        s1: PyFelt,
        Ps: list[list[ModuloCircuitElement]],
        Ps_sparsities: list[list[int] | None] = None,
        acc_index: int = 0,
    ):
        # Sanity checks for sparsities
        Ps_sparsities = Ps_sparsities or [None] * len(Ps)
        assert len(Ps_sparsities) == len(
            Ps
        ), f"len(Ps_sparsities)={len(Ps_sparsities)} != len(Ps)={len(Ps)}"
        for i, sparsity in enumerate(Ps_sparsities):
            if sparsity:
                assert all(
                    Ps[i][j].value == 0 for j in range(len(Ps[i])) if sparsity[j] == 0
                )
                assert all(
                    Ps[i][j].value == 1 for j in range(len(Ps[i])) if sparsity[j] == 2
                )

        # Evaluate LHS = Π(Pi(z)) inside circuit.
        # i=0
        LHS = self.eval_poly_in_precomputed_Z(Ps[0], Ps_sparsities[0])
        LHS_current_eval = LHS
        # Keep P0(z)
        self.accumulate_poly_instructions[acc_index].Pis_of_Z.append([])
        self.accumulate_poly_instructions[acc_index].Pis_of_Z[-1].append(LHS)
        for i in range(1, len(Ps)):
            if Ps[i - 1] == Ps[i]:
                # Consecutives elements are the same : retrieve previous evaluation.
                LHS_current_eval = self.accumulate_poly_instructions[
                    acc_index
                ].Pis_of_Z[-1][-1]

                # Todo : support smarter analysis to save a few muls

            else:
                LHS_current_eval = self.eval_poly_in_precomputed_Z(
                    Ps[i], Ps_sparsities[i]
                )
            # Keep Pi(z)
            self.accumulate_poly_instructions[acc_index].Pis_of_Z[-1].append(
                LHS_current_eval
            )
            # Update LHS
            LHS = self.mul(LHS, LHS_current_eval)

        ci_XY_of_z = self.mul(s1, LHS, "ci_XY_of_z")

        LHS_acc = self.add(self.acc[acc_index].lhs, ci_XY_of_z, "LHS_acc")

        # Update LHS only.
        self.acc[acc_index] = EuclideanPolyAccumulator(
            lhs=LHS_acc,
            R=self.acc[acc_index].R,
            R_evaluated=self.acc[acc_index].R_evaluated,
        )

        return

    def update_RHS_state(
        self,
        type: AccPolyInstructionType,
        s1: ModuloCircuitElement,
        R: list[ModuloCircuitElement],
        r_sparsity: list[int] = None,
        acc_index: int = 0,
        instruction_index: int = 0,
    ):

        # Find if R_of_z is already computed in the first Pi of the next instruction.
        # In this case we avoid accumulating R as polynomial and can use directly R_of_z with the correct s1.
        if type != AccPolyInstructionType.SQUARE_TORUS:
            if instruction_index + 1 < self.accumulate_poly_instructions[acc_index].n:
                if (
                    self.accumulate_poly_instructions[acc_index].Pis[
                        instruction_index + 1
                    ][0]
                    == R
                ):
                    already_computed_R_of_z = self.accumulate_poly_instructions[
                        acc_index
                    ].Pis_of_Z[instruction_index + 1][0]

                    # Update direclty Rhs_evaluted
                    self.acc[acc_index] = EuclideanPolyAccumulator(
                        lhs=self.acc[acc_index].lhs,
                        R=self.acc[acc_index].R,
                        R_evaluated=self.add(
                            self.acc[acc_index].R_evaluated,
                            self.mul(s1, already_computed_R_of_z),
                        ),
                    )
                    return

        # If not found, computes R_acc = R_acc + s1 * R as a Polynomial inside circuit
        if r_sparsity:
            if type != AccPolyInstructionType.SQUARE_TORUS:
                # Sanity check is already done in square_torus function.
                # If the instruction is SQUARE_TORUS, then R is fully determined by its sparsity, R(x) = x.
                # Actual R value here is the SQ TORUS result.
                # See square torus function for this edge case.
                assert all(R[i].value == 0 for i in range(len(R)) if r_sparsity[i] == 0)
            R_acc = []
            for i, (r_acc, r) in enumerate(zip(self.acc[acc_index].R, R)):
                match r_sparsity[i]:
                    case 1:
                        R_acc.append(self.add(r_acc, self.mul(s1, r)))
                    case 2:
                        R_acc.append(self.add(r_acc, s1))
                    case _:
                        R_acc.append(r_acc)

        else:
            # Computes R_acc = R_acc + s1 * R without sparsity info.
            R_acc = [
                self.add(r_acc, self.mul(s1, r))
                for r_acc, r in zip(self.acc[acc_index].R, R)
            ]
        # Update accumulator state
        self.acc[acc_index] = EuclideanPolyAccumulator(
            lhs=self.acc[acc_index].lhs,
            R=R_acc,
            R_evaluated=self.acc[acc_index].R_evaluated,
        )
        return

    def finalize_circuit(
        self,
        extension_degree: int = None,
    ):
        ######### Flags #########
        extension_degree = extension_degree or self.extension_degree
        double_extension = self.accumulate_poly_instructions[1].n > 0
        acc_indexes = [0, 1] if double_extension else [0]
        #########################

        ################ Get base rlc coefficient c0 ################
        if self.hash_input:
            self.transcript.hash_limbs_multi(self.circuit_input)
        self.transcript.hash_limbs_multi(self.commitments)

        c0 = self.write_cairo_native_felt(self.field(self.transcript.s1))
        ##################################################################
        ################ Compute Qs ################
        Qs = [Polynomial([self.field.zero()]) for _ in range(2)]

        for acc_index in acc_indexes:
            self.accumulate_poly_instructions[acc_index].rlc_coeffs.append(c0)
            # Computes Q = Σ(ci * Qi)
            Qs[acc_index] = self.accumulate_poly_instructions[acc_index].Qis[0] * c0
            for i in range(1, self.accumulate_poly_instructions[acc_index].n):
                self.accumulate_poly_instructions[acc_index].rlc_coeffs.append(
                    self.mul(
                        self.accumulate_poly_instructions[acc_index].rlc_coeffs[i - 1],
                        c0,
                    )
                )
                Qs[acc_index] += (
                    self.accumulate_poly_instructions[acc_index].Qis[i]
                    * self.accumulate_poly_instructions[acc_index].rlc_coeffs[i]
                )
            Qs[acc_index] = Qs[acc_index].get_coeffs()
            # Extend Q with zeros if needed to match the minimal expected degree.
            Qs[acc_index] = Qs[acc_index] + [self.field.zero()] * (
                (acc_index + 1) * extension_degree - 1 - len(Qs[acc_index])
            )
        ##################################################################

        Q = [self.write_elements(Qs[0], WriteOps.COMMIT)]

        compute_z_up_to = max(max(len(Qs[0]), len(Qs[1])) - 1, extension_degree)
        self.big_q_len = len(Q[0])

        self.transcript.hash_limbs_multi(Q[0])
        if double_extension:
            Q.append(self.write_elements(Qs[1], WriteOps.COMMIT))
            self.transcript.hash_limbs_multi(Q[1])
            compute_z_up_to = max(compute_z_up_to, extension_degree * 2)
            self.big_q_len = self.big_q_len + len(Q[1])

        z = self.transcript.continuable_hash

        self.create_powers_of_Z(z, max_degree=compute_z_up_to)

        for acc_index in acc_indexes:
            for i in range(self.accumulate_poly_instructions[acc_index].n):
                self.update_LHS_state(
                    s1=self.accumulate_poly_instructions[acc_index].rlc_coeffs[i],
                    Ps=self.accumulate_poly_instructions[acc_index].Pis[i],
                    Ps_sparsities=self.accumulate_poly_instructions[
                        acc_index
                    ].Ps_sparsities[i],
                    acc_index=acc_index,
                )

            for i in range(self.accumulate_poly_instructions[acc_index].n):
                self.update_RHS_state(
                    type=self.accumulate_poly_instructions[acc_index].types[i],
                    s1=self.accumulate_poly_instructions[acc_index].rlc_coeffs[i],
                    R=self.accumulate_poly_instructions[acc_index].Ris[i],
                    r_sparsity=self.accumulate_poly_instructions[
                        acc_index
                    ].r_sparsities[i],
                    acc_index=acc_index,
                    instruction_index=i,
                )

            Q_of_Z = self.eval_poly_in_precomputed_Z(Q[acc_index])
            P, P_sparsity = self.write_sparse_constant_elements(
                get_irreducible_poly(
                    self.curve_id, (acc_index + 1) * extension_degree
                ).get_coeffs(),
            )
            P_of_z = self.eval_poly_in_precomputed_Z(P, P_sparsity)
            R = self.acc[acc_index].R
            R_of_Z = self.eval_poly_in_precomputed_Z(R)

            lhs = self.acc[acc_index].lhs
            rhs = self.add(
                self.mul(Q_of_Z, P_of_z),
                self.add(R_of_Z, self.acc[acc_index].R_evaluated),
            )
            assert lhs.value == rhs.value, f"{lhs.value} != {rhs.value}, {acc_index}"
            if self.compilation_mode == 0:
                self.sub_and_assert(
                    lhs, rhs, self.set_or_get_constant(self.field.zero())
                )
            else:
                eq_check = self.sub(rhs, lhs)
                self.extend_output([eq_check])
        return True

    def summarize(self):
        add_count, mul_count, assert_eq_count = self.values_segment.summarize()
        summary = {
            "circuit": self.name,
            "MULMOD": mul_count,
            "ADDMOD": add_count,
            "ASSERT_EQ": assert_eq_count,
            "POSEIDON": (
                self.transcript.permutations_count
                if self.transcript.permutations_count > 1
                else 0
            ),
            "RLC": self.accumulate_poly_instructions[0].n
            + self.accumulate_poly_instructions[1].n,
        }

        return summary

    def compile_circuit_cairo_zero(
        self,
        function_name: str = None,
        returns: dict[str] = {
            "felt*": [
                "constants_ptr",
                "add_offsets_ptr",
                "mul_offsets_ptr",
                "output_offsets_ptr",
            ],
            "felt": [
                "constants_ptr_len",
                "input_len",
                "commitments_len",
                "big_Q_len",
                "witnesses_len",
                "output_len",
                "continuous_output",
                "add_mod_n",
                "mul_mod_n",
                "n_assert_eq",
                "N_Euclidean_equations",
                "name",
                "curve_id",
            ],
        },
    ) -> str:
        dw_arrays = self.values_segment.get_dw_lookups()
        name = function_name or self.values_segment.name
        function_name = f"get_{name}_circuit"
        code = f"func {function_name}()->(circuit:{self.class_name}*)" + "{" + "\n"

        code += "alloc_locals;\n"
        code += "let (__fp__, _) = get_fp_and_pc();\n"

        for dw_array_name in returns["felt*"]:
            code += f"let ({dw_array_name}:felt*) = get_label_location({dw_array_name}_loc);\n"

        code += f"let constants_ptr_len = {len(dw_arrays['constants_ptr'])};\n"
        code += f"let input_len = {len(self.values_segment.segment_stacks[WriteOps.INPUT])*N_LIMBS};\n"
        code += f"let commitments_len = {len(self.commitments)*N_LIMBS};\n"
        code += f"let witnesses_len = {len(self.values_segment.segment_stacks[WriteOps.WITNESS])*N_LIMBS};\n"
        code += f"let big_Q_len = {self.big_q_len*N_LIMBS};\n"
        code += f"let output_len = {len(self.output)*N_LIMBS};\n"
        continuous_output = self.continuous_output
        code += f"let continuous_output = {1 if continuous_output else 0};\n"
        code += f"let add_mod_n = {len(dw_arrays['add_offsets_ptr'])};\n"
        code += f"let mul_mod_n = {len(dw_arrays['mul_offsets_ptr'])};\n"
        code += (
            f"let n_assert_eq = {len(self.values_segment.assert_eq_instructions)};\n"
        )
        code += (
            f"let N_Euclidean_equations = {len(dw_arrays['poseidon_indexes_ptr'])};\n"
        )
        code += f"let name = '{self.name}';\n"
        code += f"let curve_id = {self.curve_id};\n"

        code += f"local circuit:ExtensionFieldModuloCircuit = ExtensionFieldModuloCircuit({', '.join(returns['felt*'])}, {', '.join(returns['felt'])});\n"
        code += "return (&circuit,);\n"

        for dw_array_name in returns["felt*"]:
            dw_values = dw_arrays[dw_array_name]
            code += f"\t {dw_array_name}_loc:\n"
            if dw_array_name == "constants_ptr":
                for bigint in dw_values:
                    for limb in bigint:
                        code += f"\t dw {limb};\n"
                code += "\n"

            elif dw_array_name in ["add_offsets_ptr", "mul_offsets_ptr"]:
                num_instructions = len(dw_values)
                instructions_needed = (
                    BATCH_SIZE - (num_instructions % BATCH_SIZE)
                ) % BATCH_SIZE
                for left, right, result, comment in dw_values:
                    code += (
                        f"\t dw {left}; // {comment}\n"
                        + f"\t dw {right};\n"
                        + f"\t dw {result};\n"
                    )
                if instructions_needed > 0:
                    first_triplet = dw_values[0]
                    for _ in range(instructions_needed):
                        code += (
                            f"\t dw {first_triplet[0]};\n"
                            + f"\t dw {first_triplet[1]};\n"
                            + f"\t dw {first_triplet[2]};\n"
                        )
                code += "\n"
            elif dw_array_name in ["output_offsets_ptr"]:
                if continuous_output:
                    code += f"\t dw {dw_values[0]};\n"
                else:
                    for val in dw_values:
                        code += f"\t dw {val};\n"

        code += "\n"
        code += "}\n"
        return code


if __name__ == "__main__":
    pass
